import click

from sym.flow.cli.helpers.api_operations import (
    Operation,
    OperationHelper,
    OperationSets,
    OperationType,
)
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.models import CSVMatcher, Identity, Service, User
from sym.flow.cli.models.service import SYM_CLOUD_KEY


def update_identity(
    options: GlobalOptions,
    user_to_update: User,
    new_identity_value: str,
    service: Service,
):
    """Given a user and a service, executes the operation helper for creating/updating the identity."""
    new_identity = Identity(
        service=service,
        matcher=CSVMatcher(service=service, value=new_identity_value).to_dict(),
    )
    sym_identity = user_to_update.get_identity_from_key(SYM_CLOUD_KEY)

    operations = OperationSets(
        # Don't need to send in all the existing identities,
        # just the identity to update and the Sym identity to identify the user
        update_user_ops=[
            Operation(
                operation_type=OperationType.update_user,
                original_value=None,
                new_value=User(
                    id=user_to_update.id, identities=[new_identity, sym_identity]
                ),
            )
        ],
        delete_identities_ops=[],
        delete_user_ops=[],
    )

    operation_helper = OperationHelper(options, operations)
    operation_helper.handle_update_users()


def delete_identity(options: GlobalOptions, original_user: User, service: Service):
    """Given a user and a service, executes the operation helper for deleting the identity."""
    identity_to_delete = original_user.get_identity_from_key(service.service_key)
    if not identity_to_delete:
        click.secho(
            f"{original_user.sym_identifier} has no identity for service {service.service_key}",
            fg="yellow",
        )

    new_identities = [i for i in original_user.identities if i != identity_to_delete]

    operations = OperationSets(
        update_user_ops=[],
        delete_identities_ops=[
            Operation(
                operation_type=OperationType.delete_identity,
                original_value=original_user,
                new_value=User(id=original_user.id, identities=new_identities),
            )
        ],
        delete_user_ops=[],
    )

    operation_helper = OperationHelper(options, operations)
    operation_helper.handle_delete_identities()



import re
from datetime import timedelta
from typing import Optional

import click

from sym.flow.cli.errors import InvalidExpiryError
from sym.flow.cli.helpers.global_options import GlobalOptions


def to_timedelta(value: str) -> int:
    """Parses the shorthand expiry and returns a timedelta if it is in an accepted format"""

    # Accepted expiries are an integer (non-zero, non-zero-leading) followed by [s, m, d, or mo]
    expiry_pattern = re.compile(r"^([1-9][0-9]*)(\w{1,2})$")
    value = value.strip()

    if expiry_value := expiry_pattern.match(value):
        duration_value = int(expiry_value.group(1))
        duration_unit = expiry_value.group(2)
        if duration_unit == "s":
            return int(timedelta(seconds=duration_value).total_seconds())
        elif duration_unit == "m":
            return int(timedelta(minutes=duration_value).total_seconds())
        elif duration_unit == "d":
            return int(timedelta(days=duration_value).total_seconds())
        elif duration_unit == "mo":
            return int(timedelta(days=30 * duration_value).total_seconds())

    raise InvalidExpiryError(value)


@click.command(name="issue", short_help="Issue a new Sym Token")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.option(
    "-u",
    "--username",
    help="The name of the bot-user to issue the token for",
    prompt="Bot username",
    required=True,
)
@click.option(
    "-e",
    "--expiry",
    help="The TTL for the token in shorthand",
    prompt=True,
    required=True,
)
@click.option(
    "-l",
    "--label",
    help="An optional label for the token",
    required=False,
)
@click.option(
    "-n",
    "--no-newline",
    is_flag=True,
    show_default=True,
    default=False,
    help="Do not print a trailing newline character",
)
def tokens_issue(
    options: GlobalOptions,
    username: str,
    expiry: str,
    no_newline: bool,
    label: Optional[str],
) -> None:
    """
    Issues a new token for the bot-user identified by the given username.
    The expiry is a time duration in shorthand, e.g. `30d`

    \b
    Accepted durations are:
        - `s` = seconds
        - `m` = minutes
        - `d` = days
        - `mo` = months (30 days)
    """
    expiry = to_timedelta(expiry)
    access_token = options.sym_api.create_token(username.strip(), expiry, label)
    newline = not no_newline
    click.secho(f"{access_token}", fg="green", nl=newline)



from unittest.mock import patch

from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.factories.resources import TerraformResourceFactory


class TestResourcesList:
    @patch("sym.flow.cli.helpers.api.SymAPI.get_resources", return_value=[])
    def test_list_no_resources_exits(self, _, click_setup):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["resources", "list", "sym_flow"])

        assert result.exit_code == 0
        assert "No sym_flow resources were found." in result.output

    def test_list_invalid_type_errors(self, click_setup):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["resources", "list", "doggo"])

        assert result.exit_code != 0
        assert "Invalid input: 'doggo'" in result.output
        assert "sym_flow, sym_target" in result.output

    def test_list_resources_with_type(self, click_setup):
        resources = [
            TerraformResourceFactory.create(),
            TerraformResourceFactory.create(type="permission_context"),
            TerraformResourceFactory.create(),
            TerraformResourceFactory.create(type="pagerduty"),
        ]

        with patch(
            "sym.flow.cli.helpers.api.SymAPI.get_resources", return_value=resources
        ):
            with click_setup() as runner:
                result = runner.invoke(
                    click_command, ["resources", "list", "sym_integration"]
                )

        assert result.exit_code == 0
        assert "Type\n" in result.output  # Ensure the table header exists
        assert "pagerduty" in result.output  # Ensure row includes type value
        for resource in resources:
            assert resource.srn in result.output

    def test_list_resources_without_type(self, click_setup):
        resources = TerraformResourceFactory.create_batch(5)
        for r in resources:
            r.sub_type = None

        with patch(
            "sym.flow.cli.helpers.api.SymAPI.get_resources", return_value=resources
        ):
            with click_setup() as runner:
                result = runner.invoke(click_command, ["resources", "list", "sym_flow"])

        assert result.exit_code == 0
        assert "Type" not in result.output  # Ensure the table header does not exist
        assert resources[4].slug in result.output
        for resource in resources:
            assert resource.srn in result.output



"""Config Get

Retrieve a value from the Sym Flow config.
"""


import click

import sym.flow.cli.helpers.cli as cli_helpers
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.global_options import GlobalOptions


@click.command(name="get", short_help="Get a config value")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.argument("key")
def config_get(options: GlobalOptions, key: str) -> None:
    """Get a config value from your local Sym Flow config file.
    Nested configs can be accessed by a dot-separated path
    e.g. `symflow config get auth_token.access_token`
    """
    # For internal use only
    try:
        value = Config.get_value(key)
    except ValueError:
        cli_helpers.fail(f"The path '{key}' is incomplete", "Please enter a full path")

    if not value:
        cli_helpers.fail(
            f"Failed to get config value for '{key}'", "The key doesn't exist"
        )

    click.echo(value)



import click

from ...helpers.debug import do_upload_directory


@click.command(short_help="Share a directory with Sym")
@click.argument("dir", type=click.Path(exists=True, file_okay=False))
def upload_directory(dir: click.Path) -> None:
    """Share the contents of DIR with Sym."""

    click.echo("Archiving and uploading...")
    txid = do_upload_directory(str(dir))
    click.secho(f"Your directory was shared with Sym!", fg="green")
    click.secho(f"Transaction ID: {txid}", fg="black", bold=True)



from dataclasses import dataclass
from typing import TypedDict

from sym.flow.cli.errors import InvalidTokenError


@dataclass
class UserCredentials:
    username: str
    password: str


class AuthToken(TypedDict):
    access_token: str
    refresh_token: str
    token_type: str
    expires_in: int
    scope: str

    def __init__(
        self,
        access_token: str,
        refresh_token: str,
        token_type: str,
        expires_in: int,
        scope: str,
    ):
        # This is necessary instead of @dataclass due to an
        # incompatibility between dataclass + TypedDict in python3.9
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.token_type = token_type
        self.expires_in = expires_in
        self.scope = scope


def parse_auth_token(json_response: dict) -> AuthToken:
    try:
        return AuthToken(
            access_token=json_response["access_token"],
            refresh_token=json_response["refresh_token"],
            token_type=json_response["token_type"],
            expires_in=json_response["expires_in"],
            scope=json_response["scope"],
        )
    except KeyError:
        raise InvalidTokenError(str(json_response))



from collections import Counter
from typing import List

import click
from tabulate import tabulate

from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.utils import human_friendly, utc_to_local
from sym.flow.cli.models import BaseService, ServiceType, User
from sym.flow.cli.models.user_type import UserType


@click.command(name="list", short_help="List all Sym Bot Users")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def bots_list(options: GlobalOptions) -> None:
    """
    Lists all bot-users and their token counts in your organization.
    """
    table_data = get_bot_users_data(options.sym_api)
    click.echo(tabulate(table_data, headers="firstrow"))


def get_bot_users_data(api: SymAPI) -> List[List[str]]:
    bot_users = api.get_users({"type": UserType.BOT})
    services = _extract_services_from_user_data(bot_users)

    token_counts = Counter([str(t.user.id) for t in api.get_tokens()])
    headers = ["Username", "Token Count", "Created At"] + [
        service.service_key for service in services
    ]
    table_data = [headers]
    for u in bot_users:
        user_info = [
            u.sym_identifier,
            token_counts.get(u.id, 0),
            human_friendly(utc_to_local(u.created_at)),
        ]
        for service in services:
            user_info.append(u.identity_repr_for_service(service))
        table_data.append(user_info)

    return table_data


def _extract_services_from_user_data(user_data: List[User]) -> List[BaseService]:
    services = []
    for user in user_data:
        for identity in user.identities:
            if identity.service.service_type != ServiceType.SYM.type_name:
                services.append(identity.service)

    return services



from typing import Optional

from auth0.v3.authentication.token_verifier import (
    AsymmetricSignatureVerifier,
    TokenVerifier,
)
from pydantic import BaseModel, Field


class JWT(BaseModel):
    class Config:
        allow_population_by_field_name = True

    # Used to map to an Organization
    client_id: str = Field(alias="azp")

    # The type of the connection in Auth0
    connection_strategy: str = Field(alias="https://api.symops.com/connection_strategy")

    # The name of the connection in Auth0
    connection_name: str = Field(alias="https://api.symops.com/connection_name")

    # Email & Name of the user associated with the JWT
    email: str = Field(alias="https://api.symops.com/email")
    name: str = Field(alias="https://api.symops.com/name")

    # The ID of the user in the underlying IDP
    idp_id: Optional[str] = Field(alias="https://api.symops.com/idp_id", default=None)

    @classmethod
    def from_access_token(cls, access_token: str) -> "JWT":
        """Decode the provided access token and return a JWT object."""
        verifier = TokenVerifier(
            signature_verifier=AsymmetricSignatureVerifier(
                jwks_url="https://symops.us.auth0.com/.well-known/jwks.json",
                algorithm="RS256",
            ),
            issuer="https://symops.us.auth0.com/",
            audience="https://api.symops.com",
        )

        decoded_token = verifier.verify(access_token)
        return cls.parse_obj(decoded_token)



from unittest.mock import patch

from sym.flow.cli.symflow import symflow as click_command


class TestBotsDelete:
    @patch("sym.flow.cli.helpers.api.SymAPI.delete_user")
    def test_delete(
        self,
        mock_delete_user,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["bots", "delete", "bot-user"])
            assert result.exit_code == 0
            assert "Successfully deleted bot user bot-user!" in result.output

        mock_delete_user.assert_called_once()

    @patch("sym.flow.cli.helpers.api.SymAPI.delete_user")
    def test_delete_missing_arguments(
        self,
        mock_delete_user,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["bots", "delete"])
            assert result.exit_code == 2

        mock_delete_user.assert_not_called()



import click
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.commands.bots.create import bots_create
from sym.flow.cli.commands.bots.delete import bots_delete
from sym.flow.cli.commands.bots.delete_identity import delete_bot_identity
from sym.flow.cli.commands.bots.list import bots_list
from sym.flow.cli.commands.bots.update_identity import update_bot_identity
from sym.flow.cli.helpers.global_options import GlobalOptions


@click.group(name="bots", cls=SymGroup, short_help="Perform operations on Sym Bot Users")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def bots(options: GlobalOptions) -> None:
    """Operations on Sym Bot Users"""


bots.add_command(bots_create)
bots.add_command(bots_list)
bots.add_command(bots_delete)
bots.add_command(update_bot_identity)
bots.add_command(delete_bot_identity)



from unittest.mock import patch

import pytest

from sym.flow.cli.helpers.users import UserUpdateSet
from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.factories.users import UserFactory
from sym.flow.cli.tests.helpers.sandbox import Sandbox


@patch("sym.flow.cli.commands.users.update.click.edit")
class TestUsersUpdate:
    @pytest.fixture
    def users(self):
        yield UserFactory.create_batch(2)

    @pytest.fixture
    def services(self, users):
        yield [i.service for i in users[0].identities]

    @pytest.fixture(autouse=True)
    def fixture_patch(self, users, services):
        """Set up class-level patches using fixtures, since they can't be passed
        directly into patch decorators.
        """
        with patch("sym.flow.cli.helpers.api.SymAPI.get_users", return_value=users):
            with patch(
                "sym.flow.cli.helpers.api.SymAPI.get_services", return_value=services
            ):
                yield

    @patch("sym.flow.cli.commands.users.update.click.confirm", return_value=True)
    @patch("sym.flow.cli.helpers.api_operations.OperationHelper.apply_changes")
    def test_click_calls_execution_method(
        self,
        mock_operation_apply_changes,
        _,
        mock_click_edit,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["users", "update"])
            assert result.exit_code == 0

        mock_click_edit.assert_called_once()
        mock_operation_apply_changes.assert_called_once()

    @patch("sym.flow.cli.helpers.api_operations.OperationHelper.apply_changes")
    def test_no_confirm_does_not_edit(
        self,
        mock_operation_apply_changes,
        mock_click_edit,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["users", "update"])
            assert result.exit_code == 1

        mock_click_edit.assert_not_called()
        mock_operation_apply_changes.assert_not_called()

    @patch("sym.flow.cli.helpers.api_operations.OperationHelper.apply_changes")
    def test_file_input(
        self,
        mock_operation_apply_changes,
        mock_click_edit,
        click_setup,
        users,
        services,
        sandbox: Sandbox,
    ):
        filepath = f"{sandbox.path}/pytest.csv"
        UserUpdateSet(user_data=users, service_data=services).write_to_csv(filepath)

        with click_setup() as runner:
            result = runner.invoke(
                click_command, ["users", "update", "--input-file", filepath]
            )
            assert result.exit_code == 0

        mock_click_edit.assert_not_called()
        mock_operation_apply_changes.assert_called_once()

    @patch("sym.flow.cli.commands.users.update.click.confirm", return_value=True)
    @patch("sym.flow.cli.helpers.users.UserUpdateSet.add_csv_row")
    @patch("sym.flow.cli.helpers.api_operations.OperationHelper.apply_changes")
    def test_edit_the_correct_amount_of_users(
        self,
        _,
        mock_add_csv_row,
        __,
        ___,
        click_setup,
        users,
    ):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["users", "update"])
            assert result.exit_code == 0

        assert mock_add_csv_row.call_count == len(users)



import json
from datetime import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel, Field
from sym.shared.cli.errors import UnknownError

from sym.flow.cli.models.service import SERVICE_KEY_SEPARATOR, BaseService
from sym.flow.cli.models.service_type import ServiceType
from sym.flow.cli.models.user_type import UserType


class CSVMatcher(BaseModel):
    service: BaseService
    value: str = Field(default_value="")

    @classmethod
    def from_dict(cls, service: BaseService, matcher: Dict):
        if service_type := ServiceType.get(service.service_type):
            value = matcher[service_type.matcher]
        else:
            value = json.dumps(matcher)
        return cls(service=service, value=value)

    @property
    def default_key(self) -> Optional[str]:
        if service_type := ServiceType.get(self.service.service_type):
            return service_type.matcher

        return None

    def to_dict(self) -> Dict:
        """Tries to generate the matcher dict if the service matcher key
        is known, otherwise use the value directly.
        """
        if self.default_key:
            return {self.default_key: self.value}

        try:
            return json.loads(self.value)
        except ValueError:
            raise ValueError(
                "Provided identity matcher '{}' for service '{}' should be valid json.".format(
                    self.value, self.service
                )
            )


class Identity(BaseModel):
    service: BaseService
    matcher: dict

    @classmethod
    def from_csv(cls, service_key: str, matcher_value: str) -> "Identity":
        service_type, external_id = Identity.parse_service_key(service_key=service_key)
        service = BaseService(slug=service_type, external_id=external_id)
        matcher = CSVMatcher(service=service, value=matcher_value).to_dict()
        return cls(service=service, matcher=matcher)

    @classmethod
    def parse_service_key(cls, service_key: str) -> List[str]:
        return service_key.split(SERVICE_KEY_SEPARATOR, 1)

    def to_csv(self) -> str:
        return CSVMatcher.from_dict(self.service, self.matcher).value


class User(BaseModel):
    id: str
    identities: List[Identity] = []
    type: UserType = UserType.NORMAL
    # Optional because this is only used to parse from the API, not when creating users
    created_at: Optional[datetime]
    is_admin: bool = False

    @property
    def sym_email(self) -> str:
        for identity in self.identities:
            if identity.service.service_type == ServiceType.SYM.type_name:
                return identity.matcher["email"]

        raise UnknownError(Exception(f"No Sym Identity found for user id {self.id}"))

    @property
    def sym_identifier(self) -> str:
        """Returns value in the SYM Identity matcher, regardless of email or username"""
        if identifier := next(
            (
                list(identity.matcher.values())[
                    0
                ]  # Assumption is that Sym identity matcher has only one value
                for identity in self.identities
                if identity.service.service_type == ServiceType.SYM.type_name
            ),
            None,
        ):
            return identifier

        raise UnknownError(Exception(f"No Sym Identify found for user id {self.id}"))

    @property
    def identities_without_sym_service(self) -> List[Identity]:
        return self.filtered_identities([ServiceType.SYM.type_name])

    def filtered_identities(self, excluded_service_types: List[str]) -> List[Identity]:
        return [
            identity
            for identity in self.identities
            if identity.service.service_type not in excluded_service_types
        ]

    def get_identity_from_key(self, key: str) -> Optional[Identity]:
        service_type, external_id = Identity.parse_service_key(service_key=key)
        return next(
            iter(
                [
                    i
                    for i in self.identities
                    if i.service.service_type == service_type
                    and i.service.external_id == external_id
                ]
            ),
            None,
        )

    def identity_repr_for_service(self, service: BaseService) -> str:
        if identity := self.get_identity_from_key(service.service_key):
            return identity.to_csv()
        return ""



from sym.flow.cli.helpers.config import store_login_config
from sym.flow.cli.symflow import symflow as click_command


def test_config_get_org(click_setup, test_org, auth_token):
    with click_setup() as runner:
        store_login_config("foo@bar.com", test_org, auth_token)
        result = runner.invoke(click_command, ["config", "get", "org"])

        assert result.stdout.strip() == test_org.slug



import click
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.helpers.global_options import GlobalOptions

from .upload_directory import upload_directory
from .upload_file import upload_file
from .upload_terraform_output import upload_terraform_output


@click.group(
    cls=SymGroup,
    short_help="Utility commands for debugging a Sym Runtime",
    hidden=True,
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def debug(options: GlobalOptions) -> None:
    """Sym Flow CLI commands for debugging.

    For more details, see https://docs.symops.com/docs/support
    """


debug.add_command(upload_terraform_output)
debug.add_command(upload_file)
debug.add_command(upload_directory)



import click

from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.models.service import BaseService
from sym.flow.cli.models.user import CSVMatcher


@click.command(
    name="delete",
    short_help="Delete a User",
)
@click.argument("service", required=True, type=str)
@click.option("--user-id", required=True, type=str)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def users_delete(options: GlobalOptions, service: str, user_id: str) -> None:
    # External ID is not used, using service as a placeholder
    service_obj = BaseService(slug=service, external_id=service)
    matcher = CSVMatcher(service=service_obj, value=user_id)

    payload = {
        "users": [
            {
                "identity": {
                    "service_type": service,
                    "matcher": matcher.to_dict(),
                },
            }
        ]
    }

    options.sym_api.delete_user(payload)
    click.secho(f"Successfully deleted user {user_id}!")



import uuid
import webbrowser
from abc import ABC, abstractmethod
from http.server import HTTPServer
from typing import Tuple

import click
import pkce
import requests
from sym.shared.cli.errors import CliError

from sym.flow.cli.helpers.constants import DEFAULT_SCOPES
from sym.flow.cli.helpers.login.handler import make_handler

from ...errors import LoginError
from ...helpers.global_options import GlobalOptions
from ...models import AuthToken, Organization, UserCredentials, parse_auth_token


class LoginFlow(ABC):
    @abstractmethod
    def login(self, options: GlobalOptions, org: Organization) -> AuthToken:
        pass

    @classmethod
    def get_login_flow(cls, browser: bool, port: int) -> "LoginFlow":
        if not browser:
            return PasswordPromptFlow()
        else:
            return BrowserRedirectFlow(port)


class PasswordPromptFlow(LoginFlow):
    """
    Prompt the user for a username and password via the command line.
    Issue a request to Auth0 with the provided user credentials to acquire an
    access token and refresh token for the Sym API.

    Implements the auth flow described here:
    https://auth0.com/docs/flows/call-your-api-using-resource-owner-password-flow
    """

    def login(self, options: GlobalOptions, org: Organization) -> AuthToken:
        creds = self.prompt_for_user_credentials()
        return self.login_with_user_creds(options, creds, org)

    def prompt_for_user_credentials(self) -> UserCredentials:
        username = click.prompt("Please enter your username", type=str)
        password = click.prompt("Please enter your password", type=str, hide_input=True)
        return UserCredentials(username, password)

    def login_with_user_creds(
        self, options: GlobalOptions, creds: UserCredentials, org: Organization
    ) -> AuthToken:

        data = {
            "grant_type": "http://auth0.com/oauth/grant-type/password-realm",
            "realm": "Username-Password-Authentication",
            "username": creds.username,
            "password": creds.password,
            "audience": options.api_url,
            "scope": " ".join(DEFAULT_SCOPES),
            "client_id": org.client_id,
        }
        headers = {"content-type": "application/x-www-form-urlencoded"}
        url = f"{options.auth_url}/oauth/token"
        r = requests.post(url, headers=headers, data=data)
        if not r.ok:
            raise LoginError(r.text)

        return parse_auth_token(r.json())


class BrowserRedirectFlow(LoginFlow):
    """
    Issue an authorization code request by opening the browser to a particular
    URL. Start a server and wait for the user to successfully login. The browser will
    redirect a successful login back to the locally running server, with a request
    body containing an auth code. With the auth code, obtain an access and refresh token.

    Implements the auth code flow described here:
    https://www.altostra.com/blog/cli-authentication-with-auth0
    """

    port: int

    def __init__(self, port: int):
        self.port = port
        self.browser_login_params = None

    def login(self, options: GlobalOptions, org: Organization) -> AuthToken:
        auth_code, code_verifier = self.get_auth_code(options, org)
        if not auth_code:
            raise CliError("Unable to get access token from Auth0. Please try again.")
        return self.get_token_from_code(options, org, auth_code, code_verifier)

    @property
    def redirect_url(self) -> str:
        return f"http://localhost:{self.port}/callback"

    def _new_server(self, options: GlobalOptions, handler):
        for port in range(self.port, self.port + 10):  # try a few
            try:
                return HTTPServer(("localhost", port), handler)
            except OSError as e:
                if e.errno not in (48, 98):  # Address already in use
                    raise
                options.dprint(f"Port {port} in use, trying to increment...")

    def wait_for_code(self, state: str, options: GlobalOptions):
        """
        Starts a server and waits for a request that contains an auth code.
        """
        handler = make_handler(options, state)
        with self._new_server(options, handler) as httpd:
            options.dprint(f"Waiting for an auth code at {self.redirect_url}")
            httpd.handle_request()
            return getattr(handler, "auth_code", None)

    def gen_state(self) -> str:
        return str(uuid.uuid4())

    def gen_browser_login_params(
        self, options: GlobalOptions, org: Organization, force: bool = False
    ) -> str:
        if force or self.browser_login_params is None:
            scopes_str = " ".join(DEFAULT_SCOPES)
            state = self.gen_state()
            code_verifier = pkce.generate_code_verifier(length=128)
            code_challenge = pkce.get_code_challenge(code_verifier)
            query_params = "&".join(
                [
                    "response_type=code",
                    f"client_id={org.client_id}",
                    "code_challenge_method=S256",
                    f"code_challenge={code_challenge}",
                    f"redirect_uri={self.redirect_url}",
                    f"scope={scopes_str}",
                    f"audience=https://api.symops.com",
                    f"state={state}",
                    "prompt=login",
                ]
            )
            url = f"{options.auth_url}/authorize?{query_params}"
            self.browser_login_params = (url, state, code_verifier)

        return self.browser_login_params

    def get_auth_code(self, options: GlobalOptions, org: Organization) -> Tuple[str, str]:
        """
        Perform the Auth0 code flow by opening the browser with a URL containing query parameters
        that orchestrate the flow. When the user logs in successfully, the browser will follow the
        redirect and issue a request back to the CLI with an auth code.

        Returns an auth code and a PKCE code verifier that both need to be provided to the
        token endpoint when obtaining an auth token.
        """
        (url, state, code_verifier) = self.gen_browser_login_params(options, org)

        webbrowser.open(url)

        auth_code = self.wait_for_code(state, options)
        options.dprint(f"Auth code received: {auth_code}")
        return auth_code, code_verifier

    def get_token_from_code(
        self,
        options: GlobalOptions,
        org: Organization,
        auth_code: str,
        code_verifier: str,
    ) -> AuthToken:
        """
        With the provided auth code and verifier, obtains an auth token.
        """
        data = {
            "grant_type": "authorization_code",
            "client_id": org.client_id,
            "code_verifier": code_verifier,
            "code": auth_code,
            "redirect_uri": self.redirect_url,
        }
        headers = {"content-type": "application/x-www-form-urlencoded"}
        url = f"{options.auth_url}/oauth/token"
        r = requests.post(url, headers=headers, data=data)
        if not r.ok:
            raise LoginError(r.text)

        return parse_auth_token(r.json())



import click
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.commands.services.services_create import services_create
from sym.flow.cli.commands.services.services_delete import services_delete
from sym.flow.cli.commands.services.services_list import services_list
from sym.flow.cli.helpers.global_options import GlobalOptions


@click.group(
    name="services", cls=SymGroup, short_help="Perform operations on Sym Services"
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def services(options: GlobalOptions) -> None:
    """Operations on Services"""


services.add_command(services_list)
services.add_command(services_create)
services.add_command(services_delete)



from typing import List

import click
from tabulate import tabulate

from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.models.service import Service


@click.command(name="list", short_help="View all services")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def services_list(options: GlobalOptions) -> None:
    """View all Services currently set up for your organization."""
    click.echo(get_services_data(options.sym_api))


def get_services_data(api: SymAPI) -> str:
    services: List[Service] = api.get_services()

    table_data = [["Service Type", "External ID", "Label"]]
    for s in services:
        table_data.append([s.service_type, s.external_id, s.label])

    return tabulate(table_data, headers="firstrow")



import uuid
from typing import Dict, List, Optional

import click
import inflection
import inquirer

from sym.flow.cli.errors import InvalidChoiceError, MissingServiceError, UserAlreadyExists
from sym.flow.cli.helpers.api_operations import (
    Operation,
    OperationHelper,
    OperationSets,
    OperationType,
)
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import ServiceType
from sym.flow.cli.models.user import CSVMatcher, Identity, User


def to_service_list(ctx: click.Context, param, value) -> Dict[ServiceType, List[Service]]:
    """Parses the space separated --identities list into services to set up identities for
    e.g. "--identities slack aws_sso"

    Validates that the requested services exist in the organization
    Returns a list of services to prompt for identities
    """
    options: GlobalOptions = ctx.find_object(GlobalOptions)
    registered_services = options.sym_api.get_services()

    # always set up a sym identity
    sym_service = next(
        (s for s in registered_services if s.service_type == ServiceType.SYM.type_name)
    )
    requested_services = {ServiceType.SYM: [sym_service]}

    for service_type_str in value:
        service_type = ServiceType.get(service_type_str)
        # Requested service is not recognized
        if not service_type:
            existing_types = {s.service_type for s in registered_services}
            raise InvalidChoiceError(service_type_str, sorted(list(existing_types)))

        if services := [
            s for s in registered_services if s.service_type == str(service_type)
        ]:
            requested_services[service_type] = services
        else:
            # A valid service type, but not one registered with the organization
            raise MissingServiceError(service_type_str)

    return requested_services


def prompt_for_identity(
    service_type: ServiceType, service: Service
) -> Optional[Identity]:
    prompt = f"{inflection.titleize(service_type.type_name)} ({service.external_id}) {service_type.matcher}"

    if identity_value := inquirer.text(prompt).strip():
        return Identity(
            service=service,
            matcher=CSVMatcher(service=service, value=identity_value).to_dict(),
        )

    return None


@click.command(
    name="create",
    short_help="Create a new User",
)
@click.argument("email", required=True, type=str)
@click.option(
    "-s",
    "--service-type",
    "prompt_identities",
    multiple=True,
    callback=to_service_list,
    help="Service type to set up an Identity for",
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def users_create(
    options: GlobalOptions,
    email: str,
    prompt_identities: Dict[ServiceType, List[Service]],
) -> None:
    """Creates a single user with the given email and starts a wizard to set up identities for the service types specified

    \b
    Example:
        `symflow users create user@symops.io -s aws_sso -s slack`
    """
    users = options.sym_api.get_users()

    if email in [u.sym_email for u in users]:
        raise UserAlreadyExists(email)

    # Sym Service should always exist, set by `to_service_list`
    sym_service = prompt_identities.pop(ServiceType.SYM)[0]
    identities = [
        Identity(
            service=sym_service,
            matcher=CSVMatcher(service=sym_service, value=email).to_dict(),
        )
    ]

    if prompt_identities:
        click.echo("Leave blank to skip identity")
        # Prompt for identity values for any remaining services
        for (service_type, services) in prompt_identities.items():
            for service in services:
                if identity := prompt_for_identity(service_type, service):
                    identities.append(identity)

    operations = OperationSets(
        # update_user also handles create user
        update_user_ops=[
            Operation(
                operation_type=OperationType.update_user,
                original_value=None,
                new_value=User(id=str(uuid.uuid4()), identities=identities),
            )
        ],
        delete_identities_ops=[],
        delete_user_ops=[],
    )

    operation_helper = OperationHelper(options, operations)
    operation_helper.handle_update_users()



import sys
from typing import TYPE_CHECKING, cast

import click
import requests
import validators

from sym.flow.cli.helpers.config import Config, store_login_config
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.jwt import JWT
from sym.flow.cli.helpers.login.login_flow import LoginFlow

if TYPE_CHECKING:
    from sym.flow.cli.helpers.login.login_flow import BrowserRedirectFlow


def validate_email(ctx, param, value):
    """Returns value if the email address is valid, raises otherwise."""
    if value and not validators.email(value):
        raise click.BadParameter("must enter a valid email address")
    return value


@click.command(short_help="Log in to your Sym account")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.option(
    "--browser/--no-browser",
    default=True,
    is_flag=True,
)
@click.option(
    "--port",
    default=11001,
)
@click.option(
    "--org-slug",
    prompt=True,
    default=lambda: Config.get_org().slug if Config.is_logged_in() else None,
)
def login(options: GlobalOptions, browser: bool, port: int, org_slug: str) -> None:
    """Log in to your Sym account to authenticate with the Sym API. This is required to enable privileged actions within your Organization such as authoring Sym Flows or installing the Slack App.

    \b
    Example:
        `symflow login --org-slug sym`
    """
    if options.access_token:
        click.secho("SYM_JWT is set, unset SYM_JWT to log in as a regular user", fg="red")
        sys.exit(1)

    org = options.sym_api.get_organization_from_slug(org_slug)
    flow = LoginFlow.get_login_flow(browser, port)

    if browser:
        (url, *_) = cast("BrowserRedirectFlow", flow).gen_browser_login_params(
            options, org
        )
        styled_url = click.style(requests.utils.requote_uri(url), bold=True)  # type: ignore
        click.echo(
            f"Opening the login page in your browser. If this doesn't work, please visit the following URL:\n"
            f"{styled_url}\n"
        )
    auth_token = flow.login(options, org)
    options.set_access_token(auth_token["access_token"])
    options.dprint(auth_token=auth_token)
    click.echo("Login succeeded")

    jwt = JWT.from_access_token(auth_token["access_token"])
    store_login_config(jwt.email, org, auth_token)


@click.command(short_help="Log out of your Sym account")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def logout(options: GlobalOptions) -> None:
    if not Config.is_logged_in() and not options.access_token:
        click.secho("✖ You are already logged out!", fg="red")
        sys.exit(1)

    if options.access_token:
        click.secho(
            "✖ You are logged in via SYM_JWT, you must unset SYM_JWT manually", fg="red"
        )
        sys.exit(1)

    if Config.is_logged_in():
        Config.logout()
        click.secho("✔️  You successfully logged out!", fg="green")



import copy
from unittest.mock import patch

import pytest

from sym.flow.cli.helpers.api_operations import (
    Operation,
    OperationHelper,
    OperationSets,
    OperationType,
)
from sym.flow.cli.models.user import Identity, User
from sym.flow.cli.tests.conftest import get_mock_response
from sym.flow.cli.tests.factories.users import ServiceFactory, UserFactory


class TestOperationHelper:
    @pytest.fixture
    def user(self):
        user: User = UserFactory.create()
        user.identities.append(
            Identity(
                service=ServiceFactory(slug="google", external_id="symops.io"),
                matcher={"email": "google_user@symops.io"},
            )
        )
        return user

    @pytest.fixture
    def update_user_ops(self, user):
        edited_user = copy.deepcopy(user)

        edited_user.get_identity_from_key("google:symops.io").matcher = {
            "email": "new_google_user@symops.io"
        }

        return [
            Operation(
                operation_type=OperationType.update_user,
                original_value=user,
                new_value=edited_user,
            )
        ]

    @pytest.fixture
    def delete_identities_ops(self, user):
        edited_user = copy.deepcopy(user)

        edited_user.identities = [
            ident
            for ident in edited_user.identities
            if ident.service.service_type != "google"
        ]

        return [
            Operation(
                operation_type=OperationType.delete_identity,
                original_value=user,
                new_value=edited_user,
            )
        ]

    @pytest.fixture
    def delete_user_ops(self, user):
        return [
            Operation(
                operation_type=OperationType.delete_user,
                original_value=user,
            )
        ]

    def test_update_operations(self, global_options, update_user_ops):
        operation_helper = OperationHelper(
            global_options,
            operations=OperationSets(update_user_ops=update_user_ops),
        )
        assert len(operation_helper.update_users_payload.get("users")) == 1

        update = [
            ident["matcher"]["email"]
            for ident in operation_helper.update_users_payload["users"][0]["identities"]
            if ident["service_type"] == "google"
        ]
        assert update[0] == "new_google_user@symops.io"

    def test_delete_identity_operations(self, global_options, delete_identities_ops):

        operation_helper = OperationHelper(
            global_options,
            operations=OperationSets(delete_identities_ops=delete_identities_ops),
        )
        idents = operation_helper.delete_user_identity_payload.get("identities")
        assert len(idents) == 1
        assert idents[0]["service_type"] == "google"
        assert idents[0]["external_id"] == "symops.io"

    def test_delete_user_ops(self, global_options, delete_user_ops):
        operation_helper = OperationHelper(
            global_options,
            operations=OperationSets(delete_user_ops=delete_user_ops),
        )
        idents = operation_helper.delete_users_payload.get("users")
        assert len(idents) == 1
        assert idents[0]["id"] == delete_user_ops[0].original_value.id

    @patch("sym.flow.cli.helpers.api_operations.click.confirm", return_value=True)
    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.post",
        return_value=(
            get_mock_response(
                200,
                {
                    "status": "success",
                    "errors": [],
                    "succeeded": 1,
                    "failed": 0,
                    "results": [],
                },
            )
        ),
    )
    def test_delete_user_confirm(
        self, patch_post, patch_confirm, global_options, delete_user_ops
    ):
        operation_helper = OperationHelper(
            global_options,
            operations=OperationSets(delete_user_ops=delete_user_ops),
        )
        num_deleted = operation_helper.handle_delete_users()
        assert num_deleted == 1

    @patch("sym.flow.cli.helpers.api_operations.click.confirm", return_value=False)
    def test_delete_user_no_confirm(self, patch_confirm, global_options, delete_user_ops):
        operation_helper = OperationHelper(
            global_options,
            operations=OperationSets(delete_user_ops=delete_user_ops),
        )
        num_deleted = operation_helper.handle_delete_users()
        assert num_deleted == 0



"""Config Set

Set a value in the Sym Flow config.
"""

import click

from sym.flow.cli.helpers.global_options import GlobalOptions


@click.command(name="set", short_help="Set a config value")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.argument("key")
@click.argument("value")
def config_set(options: GlobalOptions, key: str, value: str) -> None:
    """Set a config value in your local Sym Flow config file"""
    # For internal use only
    raise click.UsageError("The set operation is not currently supported.")



import click
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.commands.tokens.issue import tokens_issue
from sym.flow.cli.commands.tokens.list import tokens_list
from sym.flow.cli.commands.tokens.revoke import tokens_revoke
from sym.flow.cli.helpers.global_options import GlobalOptions


@click.group(name="tokens", cls=SymGroup, short_help="Perform operations on Sym Tokens")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def tokens(options: GlobalOptions) -> None:
    """Operations on Tokens"""


tokens.add_command(tokens_issue)
tokens.add_command(tokens_revoke)
tokens.add_command(tokens_list)



from enum import Enum


class UserType(str, Enum):
    """Enum of UserTypes that symflow can manage.
    Note: Queuer has an additional UserType.SYSTEM that is not reflected here because system users
    should not be surfaced to symflow.
    """

    NORMAL = "normal"
    BOT = "bot"



from typing import Optional

import click
import inquirer

from sym.flow.cli.commands.users.utils import get_or_prompt_service
from sym.flow.cli.errors import MissingIdentityValueError
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.identity_operations import update_identity


@click.command(
    name="update-identity", short_help="Update a User's identity for a specific service"
)
@click.argument("email", required=True, type=str)
@click.option("--new-value", "_new_value", help="The new identity value", type=str)
@click.option(
    "--service-type",
    "_service_type",
    help="The type of service the identity is tied to",
)
@click.option(
    "--external-id",
    "_external_id",
    help="The identifier for the service",
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def update_user_identity(
    options: GlobalOptions,
    email: str,
    _new_value: Optional[str],
    _service_type: Optional[str],
    _external_id: Optional[str],
) -> None:
    """For an existing Sym User, update a single Identity such as an AWS SSO Principal UUID or a Slack User ID.

    \b
    Example:
        `symflow users update-identity user@symops.io`
    """

    api = options.sym_api
    user_to_update = api.get_user(email)
    service = get_or_prompt_service(api, _service_type, _external_id)

    new_identity_value = _new_value or inquirer.text(message="New value?")

    if not new_identity_value:
        raise MissingIdentityValueError(email)

    update_identity(options, user_to_update, new_identity_value, service)



import click
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.commands.resources.list import resources_list
from sym.flow.cli.helpers.global_options import GlobalOptions


@click.group(
    name="resources", cls=SymGroup, short_help="Perform operations on Sym resources"
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def resources(options: GlobalOptions) -> None:
    """Operations on resources (e.g. sym_flow)

    Currently, only supports the `list` operation. Any other operation (e.g. `create`
    or `update`) should be performed using Terraform.

    For more information, see https://docs.symops.com/docs/terraform-provider
    """


resources.add_command(resources_list)



from typing import Optional

import click
import inquirer
from inquirer.shortcuts import confirm, list_input
from policy_sentry.querying.all import get_all_service_prefixes
from policy_sentry.querying.arns import get_arn_type_details, get_arn_types_for_service
from policy_sentry.shared.iam_data import get_service_prefix_data
from prompt_toolkit.completion import FuzzyWordCompleter
from prompt_toolkit.shortcuts import prompt

from sym.flow.cli.helpers.arn import RawARN
from sym.flow.cli.helpers.iam_policies import IAMPolicies
from sym.flow.cli.helpers.questions import ask

from .helpers import prompt_for_access_levels


def prompt_for_resource(policies: IAMPolicies, service):
    choices = get_arn_types_for_service(service).keys()
    resource = list_input(message="Select a resource to add", choices=choices)

    raw_arn = get_arn_type_details(service, resource)["raw_arn"]
    arn = RawARN(raw_arn)

    answers = ask(
        [
            inquirer.Text(name=field, message=f"Enter the {field}", default=default)
            for field, default in arn.values().items()
        ]
    )
    arn.update(answers)

    click.secho(f"Added {arn}\n", fg="green")
    policies.add_arn(str(arn))


@click.command(
    name="wizard",
    short_help="Interactively create an AWS IAM Policy",
)
@click.option("--service", help="The AWS Service to generate a policy for")
def policies_wizard(service: Optional[str]):
    services = get_all_service_prefixes()
    service_meta = {s: get_service_prefix_data(s)["service_name"] for s in services}
    completer = FuzzyWordCompleter(words=services, meta_dict=service_meta)
    service = prompt(
        "Enter an AWS Service: ",
        completer=completer,
        complete_while_typing=True,
        default=service or "",
    )

    levels = prompt_for_access_levels(service)

    policies = IAMPolicies()

    prompt_for_resource(policies, service)
    while confirm("Would you like to add another resource to the policy?"):
        prompt_for_resource(policies, service)

    policy = policies.policy_for(*levels)
    click.echo_via_pager(policy)



import click
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.commands.users.create import users_create
from sym.flow.cli.commands.users.delete import users_delete
from sym.flow.cli.commands.users.delete_identity import delete_user_identity
from sym.flow.cli.commands.users.list import users_list
from sym.flow.cli.commands.users.list_identities import users_list_identities
from sym.flow.cli.commands.users.update import users_update
from sym.flow.cli.commands.users.update_identity import update_user_identity
from sym.flow.cli.helpers.global_options import GlobalOptions


@click.group(name="users", cls=SymGroup, short_help="Perform operations on Sym Users")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def users(options: GlobalOptions) -> None:
    """Operations on Users"""


users.add_command(users_create)
users.add_command(users_list)
users.add_command(users_list_identities)
users.add_command(users_update)
users.add_command(users_delete)
users.add_command(update_user_identity)
users.add_command(delete_user_identity)



from sym.shared.cli.helpers.config import init as config_init

config_init("symflow")

from sym.shared.cli.helpers.init import sym_init

from .commands import import_all
from .commands.symflow import symflow

import_all()
sym_init()

if __name__ == "__main__":
    symflow()



import click
from tabulate import tabulate

from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.users import UserUpdateSet


@click.command(
    name="list-identities",
    short_help="List all Users' Identities",
)
@click.option(
    "-o",
    "--output-file",
    type=click.Path(exists=False),
    help="Save results to a CSV file",
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def users_list_identities(options: GlobalOptions, output_file: str) -> None:
    """Prints a table view of Sym Users in your Organization and their corresponding identities to STDOUT.
    Use the --output-file option to save the results to a file in CSV format.

    To modify users, use `symflow users update`.
    """
    user_data = get_user_data(options.sym_api)
    if output_file:
        user_data.write_to_csv(output_file)
        click.echo(f"Saved {len(user_data.users)} users to {output_file}.")
    else:
        click.echo(tabulate_user_data(user_data))


def tabulate_user_data(user_data: UserUpdateSet) -> str:
    headers = user_data.headers
    headers.remove("User ID")
    return tabulate(user_data.tabulate(), headers=headers)


def get_user_data(api: SymAPI) -> UserUpdateSet:
    return UserUpdateSet(user_data=api.get_users(), service_data=api.get_services())



from dataclasses import dataclass, field
from enum import Enum, auto
from functools import cached_property
from typing import Dict, List, Optional

import click

from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.users import User
from sym.flow.cli.models.service import SYM_CLOUD_KEY


class OperationType(Enum):
    update_user = auto()
    delete_user = auto()
    delete_identity = auto()


@dataclass
class Operation:
    operation_type: OperationType
    original_value: Optional[User] = None
    new_value: Optional[User] = None


@dataclass
class OperationSets:
    update_user_ops: List[Operation] = field(default_factory=list)
    delete_identities_ops: List[Operation] = field(default_factory=list)
    delete_user_ops: List[Operation] = field(default_factory=list)


class OperationHelper:
    def __init__(self, options: GlobalOptions, operations: OperationSets):
        self.api = options.sym_api
        self.operations = operations

    @cached_property
    def update_users_payload(self) -> Dict:
        users = []
        for operation in self.operations.update_user_ops:
            patch_identities = []

            for identity in operation.new_value.identities:
                patch_identities.append(
                    {
                        "service_type": identity.service.service_type,
                        "external_id": identity.service.external_id,
                        "matcher": identity.matcher,
                    }
                )
            users.append(
                {
                    "id": operation.new_value.id,
                    "identities": patch_identities,
                    "type": operation.new_value.type,
                }
            )
        return {"users": users}

    @cached_property
    def delete_user_identity_payload(self) -> Dict:
        identities = []

        for operation in self.operations.delete_identities_ops:
            identities_to_delete = set(
                [
                    identity.service.service_key
                    for identity in operation.original_value.identities_without_sym_service
                ]
            ) - set(
                [
                    identity.service.service_key
                    for identity in operation.new_value.identities
                ]
            )

            for identity in operation.original_value.identities_without_sym_service:
                if identity.service.service_key in identities_to_delete:
                    identities.append(
                        {
                            "user_id": operation.new_value.id,
                            "service_type": identity.service.service_type,
                            "external_id": identity.service.external_id,
                            "matcher": identity.matcher,
                        }
                    )

        return {"identities": identities}

    @cached_property
    def delete_users_payload(self) -> Dict:
        users = []
        for operation in self.operations.delete_user_ops:
            if not (
                sym_service_identity := operation.original_value.get_identity_from_key(
                    SYM_CLOUD_KEY
                )
            ):
                continue
            users.append(
                {
                    "id": operation.original_value.id,
                    "identity": {
                        "service_type": sym_service_identity.service.service_type,
                        "matcher": sym_service_identity.matcher,
                    },
                }
            )
        return {"users": users}

    def handle_update_users(self) -> int:
        if not self.update_users_payload["users"]:
            return 0
        res = self.api.update_users(self.update_users_payload)
        update_count = res["succeeded"]
        click.secho(f"Successfully updated {update_count} users!")
        return update_count

    def handle_delete_identities(self) -> int:
        if not self.delete_user_identity_payload["identities"]:
            return 0
        res = self.api.delete_identities(self.delete_user_identity_payload)
        delete_count = res["succeeded"]
        click.secho(f"Successfully deleted {delete_count} identities!")
        return delete_count

    def handle_delete_users(self) -> int:
        if (deleted_dict := self.delete_users_payload["users"]) and click.confirm(
            f"About to delete {len(deleted_dict)} users. Do you want to continue?"
        ):
            res = self.api.delete_user(self.delete_users_payload)
            delete_count = res["succeeded"]
            click.secho(f"Successfully deleted {delete_count} users!")
            return delete_count

        return 0

    def apply_changes(self):
        total_updates = 0
        total_updates += self.handle_update_users()
        total_updates += self.handle_delete_identities()
        total_updates += self.handle_delete_users()

        if not total_updates:
            click.secho(f"No changes detected!")



from unittest.mock import patch

import pytest

from sym.flow.cli.commands.bots.list import get_bot_users_data
from sym.flow.cli.helpers.utils import human_friendly, utc_to_local
from sym.flow.cli.models import ServiceType
from sym.flow.cli.tests.factories.tokens import SymTokenFactory
from sym.flow.cli.tests.factories.users import BotUserFactory, IdentityFactory


class TestBotsList:
    @pytest.fixture
    def tokens_and_bots(self):
        bot_users = BotUserFactory.create_batch(3)
        tokens = [
            SymTokenFactory.create(user__id=bot_users[0].id),
            SymTokenFactory.create(user__id=bot_users[1].id),
        ]
        yield (tokens, bot_users)

    @pytest.fixture(autouse=True)
    def patchypatch(self, tokens_and_bots):
        tokens, bots = tokens_and_bots
        with patch("sym.flow.cli.helpers.api.SymAPI.get_users", return_value=bots):
            with patch("sym.flow.cli.helpers.api.SymAPI.get_tokens", return_value=tokens):
                yield

    def test_get_bot_users_data(self, tokens_and_bots, global_options):
        tokens, bots = tokens_and_bots
        table_data = get_bot_users_data(global_options.sym_api)

        assert table_data.pop(0) == ["Username", "Token Count", "Created At"]
        for i, row in enumerate(table_data):
            assert row == [
                bots[i].sym_identifier,
                1 if i == 0 or i == 1 else 0,  # Bot users 0 and 1 have 1 token
                human_friendly(utc_to_local(bots[i].created_at)),
            ]

    def test_get_bot_users_data_with_identity(self, tokens_and_bots, global_options):
        tokens, bots = tokens_and_bots
        bots[0].identities.append(
            IdentityFactory(
                service__slug=ServiceType.PAGERDUTY.type_name,
                service__external_id="pytest-pd",
                matcher={"user_id": "PD123"},
            )
        )
        table_data = get_bot_users_data(global_options.sym_api)

        assert table_data.pop(0) == [
            "Username",
            "Token Count",
            "Created At",
            "pagerduty:pytest-pd",
        ]
        for i, row in enumerate(table_data):
            identity = bots[i].get_identity_from_key("pagerduty:pytest-pd")

            assert row == [
                bots[i].sym_identifier,
                1 if i == 0 or i == 1 else 0,  # Bot users 0 and 1 have 1 token
                human_friendly(utc_to_local(bots[i].created_at)),
                identity.to_csv() if identity else "",
            ]



import click

from sym.flow.cli.helpers.global_options import GlobalOptions

from ...helpers.debug import do_upload_terraform_output


@click.command(short_help="Share Terraform output with Sym")
@click.argument("tf_dir", type=click.Path(exists=True, file_okay=False), default=".")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def upload_terraform_output(options: GlobalOptions, tf_dir: click.Path) -> None:
    """Share the output from Terraform in TF_DIR with Sym."""

    click.echo("Running `terraform output` and uploading...")
    txid = do_upload_terraform_output(str(tf_dir), options)
    click.secho(f"Your Terraform output was shared with Sym!", fg="green")
    click.secho(f"Transaction ID: {txid}", fg="black", bold=True)



from unittest.mock import patch

import pytest

from sym.flow.cli.errors import NotAuthorizedError
from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import MANAGED_SERVICES, ServiceType
from sym.flow.cli.symflow import symflow as click_command


class TestServicesCreate:

    EXTERNAL_ID_INPUTS = {
        ServiceType.SLACK: "T123ABC",
        ServiceType.AWS_SSO: "arn:aws:sso:::instance/",
        ServiceType.AWS_IAM: "123456789012",
    }
    BAD_EXTERNAL_ID_INPUTS = {
        ServiceType.SLACK: ["T123abc", "123ABC"],  # not all upper, doesn't start with T
        ServiceType.AWS_SSO: ["instance/"],  # invalid arn
        ServiceType.AWS_IAM: ["1234abc"],  # not 12 digits
    }

    MOCK_SLACK_SERVICE = Service(
        id="fake-uuid", slug="slack", external_id="T123ABC", label="Slack"
    )

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.create_service", return_value=MOCK_SLACK_SERVICE
    )
    @patch("sym.flow.cli.commands.services.services_create.post_service_create_hooks")
    def test_services_create(
        self, mock_post_service_create_hooks, mock_create_service, click_setup
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "services",
                    "create",
                    "--service-type",
                    self.MOCK_SLACK_SERVICE.service_type,
                    "--external-id",
                    self.MOCK_SLACK_SERVICE.external_id,
                ],
            )
            assert result.exit_code == 0
            assert (
                f"Successfully set up service type slack with external ID {self.MOCK_SLACK_SERVICE.external_id}!"
                in result.output
            )
            mock_create_service.assert_called_once()
            mock_post_service_create_hooks.assert_called_once()
            mock_post_service_create_hooks.call_args.args[1] == self.MOCK_SLACK_SERVICE

    @pytest.mark.parametrize(
        "service_type",
        [s.value for s in ServiceType if s.type_name not in MANAGED_SERVICES],
    )
    def test_services_create_prompt_external_id(self, service_type, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["services", "create", "--service-type", service_type.type_name],
                input="^C",
            )
            # Check that the printed external ID is the one associated with the given service_type
            assert service_type.external_id_name in result.output

    @pytest.mark.parametrize(
        "service_type", [s.value for s in ServiceType if s.validator]
    )
    @patch("sym.flow.cli.helpers.api.SymAPI.create_service")
    def test_services_create_validation(
        self, mock_create_service, service_type, click_setup
    ):
        with click_setup() as runner:
            # Ensure that the validator is called if it is defined on a the service type
            # Example input must be defined in self.EXTERNAL_ID_INPUTS
            external_id = self.EXTERNAL_ID_INPUTS[service_type]
            result = runner.invoke(
                click_command,
                ["services", "create", "--service-type", service_type.type_name],
                input=external_id,
            )
            assert result.exit_code == 0

    @pytest.mark.parametrize(
        "service_type", [s.value for s in ServiceType if s.validator]
    )
    @patch("sym.flow.cli.helpers.api.SymAPI.create_service")
    def test_services_create_validation_failure(
        self, mock_create_service, service_type, click_setup
    ):
        with click_setup() as runner:
            # Ensure that the help_str is displayed on validation failure
            # Example input must be defined in self.BAD_EXTERNAL_ID_INPUTS
            for external_id in self.BAD_EXTERNAL_ID_INPUTS[service_type]:
                result = runner.invoke(
                    click_command,
                    ["services", "create", "--service-type", service_type.type_name],
                    input=external_id,
                )

                assert service_type.help_str in result.output
                assert result.exit_code != 0

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.create_service",
        side_effect=NotAuthorizedError,
    )
    def test_services_create_not_authorized_errors(
        self, mock_create_service, click_setup
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "services",
                    "create",
                    "--service-type",
                    "slack",
                    "--external-id",
                    "T123ABC",
                ],
            )
            mock_create_service.assert_called_once()
            assert result.exit_code is not 0



from unittest.mock import patch

import pytest

from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.factories.users import BotUserFactory


@patch("sym.flow.cli.helpers.api.SymAPI.update_users")
class TestBotsCreate:
    @pytest.fixture
    def bot_users(self):
        yield BotUserFactory.create_batch(3)

    @pytest.fixture(autouse=True)
    def patchypatch(self, bot_users):
        with patch("sym.flow.cli.helpers.api.SymAPI.get_users", return_value=bot_users):
            yield

    def test_create(self, mock_apply, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["bots", "create", "hal"],
            )
            assert result.exit_code == 0

    def test_create_bot_with_existing_username(self, mock_apply, click_setup, bot_users):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["bots", "create", bot_users[0].sym_identifier],
            )
            assert (
                f"A bot already exists with username: {bot_users[0].sym_identifier}"
                in result.output
            )



import urllib.parse as parse
from http.server import BaseHTTPRequestHandler
from typing import ClassVar, Type

from ...helpers.global_options import GlobalOptions


class AuthHTTPRequestHandler(BaseHTTPRequestHandler):

    options: ClassVar[GlobalOptions]
    state: ClassVar[str]
    auth_code: ClassVar[str]

    def log_message(self, format, *args):
        ...

    def do_GET(self):
        parsed = parse.urlparse(self.path)
        qs = parse.parse_qs(parsed.query)

        if not "code" in qs:
            self.send_error(401, explain="Missing code in query")
            return

        if not "state" in qs:
            self.send_error(401, explain="Missing state in query")
            return

        [returned_state] = qs["state"]
        if not returned_state == self.__class__.state:
            self.__class__.options.dprint(
                f"Invalid state. Found: {returned_state}, expected: {self.__class__.state}."
            )
            self.send_error(401, explain="Invalid state")
            return

        [returned_code] = qs["code"]

        self.send_response(301)
        self.send_header("Location", "https://static.symops.com/cli/login")
        self.end_headers()

        self.__class__.auth_code = returned_code


def make_handler(options: GlobalOptions, state: str) -> Type[AuthHTTPRequestHandler]:
    class Handler(AuthHTTPRequestHandler):
        ...

    Handler.options = options
    Handler.state = state
    return Handler



import json
from collections import defaultdict
from typing import Dict, List, Literal, Sequence

import inflection
from policy_sentry.querying.actions import get_actions_with_access_level
from policy_sentry.util.arns import get_service_from_arn
from policy_sentry.writing.sid_group import SidGroup
from sym.shared.cli.errors import CliError

from sym.flow.cli.helpers.arn import RawARN

# This class is used to support the `sym_resources` TF resource.
# Users can specify AWS resource ARNs to wrap a Sym workflow around,
# and this class can generate the appropriate read/write/admin policies
# for those resources, automatically.

ACTION_LEVELS = ["Read", "Write", "List", "Tagging", "Permissions management"]

PolicyAction = Literal["read", "write", "list", "tagging", "permissions-management"]
PolicySizeLimit = 2048


class IAMPolicies:
    @classmethod
    def actions_for_service(cls, service) -> Dict[str, List[str]]:
        return {al: get_actions_with_access_level(service, al) for al in ACTION_LEVELS}

    @classmethod
    def normalize_action(cls, action: str) -> PolicyAction:
        return inflection.parameterize(action)

    def __init__(self, name: str = ""):
        self.name = name
        self._arns: Dict[PolicyAction, List[str]] = defaultdict(list)
        self.sid_group = SidGroup()

    @property
    def arns(self) -> List[str]:
        return [arn for arns in self._arns.values() for arn in arns]

    def services(self) -> Dict[str, List[str]]:
        services = defaultdict(list)
        for arn in self.arns:
            services[get_service_from_arn(arn)].append(arn)
        return services

    def add_arn(self, arn: str, levels: List[str] = ACTION_LEVELS):
        if not arn.startswith("arn:"):
            arn = f"arn:{arn}"
        if not RawARN.validate(arn):
            raise CliError(f"Invalid ARN: {arn}")

        for action in levels:
            self._arns[self.normalize_action(action)].append(arn)

    def template(self, *actions: Sequence[PolicyAction], name: str) -> Dict[str, str]:
        template = {"mode": "crud", "name": f"{self.name}{name}"}
        for action in actions:
            normalized = self.normalize_action(action)
            template[normalized] = self._arns[normalized].copy()
        return template

    def policy_dict(self, template: dict):
        policy = self.sid_group.process_template(template)
        if len(json.dumps(policy).replace(" ", "")) > PolicySizeLimit:
            policy = self.sid_group.process_template(template, minimize=True)
        return policy

    def policy_for(self, *actions: Sequence[PolicyAction], name: str = ""):
        template = self.template(*actions, name=name)
        policy = self.policy_dict(template)
        return json.dumps(policy, indent=4)

    def read_policy(self):
        return self.policy_for("read", "list", name="ReadOnly")

    def write_policy(self):
        return self.policy_for("read", "list", "write", name="ReadWrite")

    def admin_policy(self):
        return self.policy_for(
            "read", "list", "write", "tagging", "permissions-management", name="Admin"
        )



"""CLI Helpers"""

import sys

import click


def fail(
    message="Something went wrong",
    hint="Please check the Sym documentation at https://docs.symops.com/.",
):
    """Raise a usage error with a useful message"""
    click.secho(f"✖ {message}.", fg="red", bold=True)
    click.secho(f"{hint}.", fg="cyan")

    sys.exit(1)



from typing import Dict, List, Optional
from uuid import UUID

from pydantic import parse_obj_as

from sym.flow.cli.errors import (
    SymAPIError,
    UnknownBotError,
    UnknownOrgError,
    UnknownUserError,
)
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.rest import SymRESTClient
from sym.flow.cli.helpers.utils import filter_dict
from sym.flow.cli.models import Organization
from sym.flow.cli.models.resource import ResourceType, TerraformResource
from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import ServiceType
from sym.flow.cli.models.token import SymToken
from sym.flow.cli.models.user import User
from sym.flow.cli.models.user_type import UserType


class SymAPI:
    def __init__(self, url: str, access_token: Optional[str] = None):
        self.rest = SymRESTClient(
            url=url, access_token=access_token or Config.get_access_token()
        )

    def set_access_token(self, access_token: str):
        self.rest.access_token = access_token

    def get_current_organization(self) -> Organization:
        """Get the organization for the currently authenticated user."""

        # There should always be exactly one organization, since the API returns specifically
        # the single organization associated with the JWT.
        org_data = self.rest.get("organizations").json()[0]
        return Organization.parse_obj(org_data)

    def get_organization_from_slug(self, org_slug: str) -> Organization:
        """Retrieves Organization data (e.g. Auth0 Client ID) from the Sym API given an org slug."""

        org_data = self.rest.get(f"auth/org/{org_slug}", force_auth=False).json()
        org = Organization.parse_obj(org_data)

        # auth/org/<slug> should always return client_id, but other endpoint don't, so it's
        # not enforced on the model. i.e. this should never happen, but check JUST in case.
        if not org.client_id:
            raise UnknownOrgError(slug=org_slug)

        return org

    def verify_login(self, email: Optional[str] = None) -> bool:
        """Returns True if the User's current credentials are valid as determined by the Sym API, False otherwise
        If an email is available to validate, include it in the request,
        """
        response = self.rest.get(
            "auth/login", filter_dict({"email": email}), validate=False
        )
        return response.status_code == 200

    def get_integrations(self) -> List[dict]:
        """Retrieve all Sym Integrations accessible to the currently
        authenticated user.
        """
        return self.rest.get("entities/integrations").json()

    def get_users(self, query_params: Optional[dict] = None) -> List[User]:
        """Retrieve all Sym Users accessible to the currently
        authenticated user.
        """

        response = self.rest.get("users", query_params).json()
        return parse_obj_as(List[User], response["users"])

    def get_user(self, email: str) -> User:
        """Given an email, get the whole user object and their identities"""
        users = self.get_users()
        user = next((u for u in users if u.sym_email == email), None)
        if not user:
            raise UnknownUserError(email=email)

        return user

    def get_user_by_id(self, user_id: UUID) -> Optional[User]:
        """Given a User UUID, get the whole user object and their identities"""
        users = self.get_users({"id": user_id})
        if users:
            return users[0]

        return None

    def get_bot(self, username: str) -> User:
        """Given a bot username, get the whole bot object and their identities"""
        bots = self.get_users({"type": UserType.BOT})
        bot = next((b for b in bots if b.sym_identifier == username), None)

        if not bot:
            raise UnknownBotError(username=username)

        return bot

    def update_users(self, payload: dict) -> List[dict]:
        return self.rest.patch("users", payload).json()

    def delete_user(self, payload: dict) -> List[dict]:
        return self.rest.post("users/delete", payload).json()

    def delete_identities(self, payload: dict) -> List[dict]:
        return self.rest.post("users/identities/delete", payload).json()

    def get_slack_install_url(self, service_id: str) -> str:
        """Get the URL that starts the Slack App installation flow.

        Returns:
            The URL to install the Slack App and the Sym Request ID used to
            retrieve the URL as a string.
        """

        return self.rest.get("services/slack/link", {"service_id": service_id}).json()[
            "url"
        ]

    def uninstall_slack(self, service_id: str) -> None:
        """Make a request to the Sym API to uninstall the Slack App.

        Raises SymAPIUnknownError from handle_response on failure. Otherwise,
        assume success.
        """

        self.rest.get("services/slack/uninstall", {"token": service_id}, validate=False)

    def get_services(
        self, service_types: List[str] = ServiceType.all_names()
    ) -> List[Service]:
        """Retrieve services in service_types registered to the currently
        authenticated user's organization.

        service_types defaults to all recognized service types
        """
        response = self.rest.get("services").json()
        services = parse_obj_as(List[Service], response["services"])

        return [s for s in services if s.service_type in service_types]

    def get_service(self, service_type: str, external_id: str) -> Service:
        """Retrieve a Service by the given service_type and external_id.
        Expects that the response will have exactly one Service
        """
        response = self.rest.get(
            "services", {"service_type": service_type, "external_id": external_id}
        ).json()
        services = parse_obj_as(List[Service], response["services"])
        if (num_services := len(services)) != 1:
            raise SymAPIError(
                f"Expected 1 service, but found {num_services} instead",
                "400",
                self.rest.last_request_id,
            )

        return services[0]

    def create_service(self, service_type: ServiceType, external_id: str) -> Service:
        """Creates a new Service for the currently authenticated user's organization"""
        response = self.rest.post(
            "services",
            {"service_type": service_type.type_name, "external_id": external_id},
        ).json()
        return Service.parse_obj(response["results"][0])

    def delete_service(self, service_type_name: str, external_id: str):
        """Deletes a Service from the currently authenticated user's organization"""
        return self.rest.post(
            "services/delete",
            {"service_type": service_type_name, "external_id": external_id},
        ).json()

    def get_service_references(self, service_id: str) -> Dict[str, List[str]]:
        """Gets all objects referencing a service, i.e. Identities/Integrations
        if any exist, a service cannot be deleted"""
        return self.rest.get(f"service/{service_id}/references").json()["references"]

    def create_token(
        self, username: str, expiry_in_seconds: int, label: Optional[str] = None
    ) -> str:
        """Creates a new token in the organization for the given bot-user and expiry"""
        response = self.rest.post(
            "tokens",
            filter_dict(
                {"username": username, "expiry": expiry_in_seconds, "label": label}
            ),
        ).json()

        return response["access_token"]

    def revoke_token(self, jti: str):
        return self.rest.post(
            "tokens/delete",
            {"identifier": jti},
        ).json()

    def get_tokens(self) -> List[SymToken]:
        response = self.rest.get(
            "tokens",
        ).json()
        return parse_obj_as(List[SymToken], response["tokens"])

    def get_resources(self, resource_type: ResourceType) -> List[TerraformResource]:
        """Retrieve all of an Organization's Terraform resources of a particular type."""

        response = self.rest.get(f"entities/{resource_type}").json()
        return parse_obj_as(List[TerraformResource], response)

    def add_domain(self, domain: str) -> None:
        """Add a domain to the organization's list."""
        org_id = self.get_current_organization().id
        self.rest.post(f"organizations/{org_id}/add-domain", {"domain": domain})

    def remove_domain(self, domain: str) -> None:
        """Remove a domain from the organization's list."""
        org_id = self.get_current_organization().id
        self.rest.post(f"organizations/{org_id}/remove-domain", {"domain": domain})



from typing import List

import click
from tabulate import tabulate

from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.utils import human_friendly, utc_to_local


@click.command(name="list", short_help="List all Users")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def users_list(options: GlobalOptions) -> None:
    """Lists all Users in your organization.

    Note: To see a list of all User identities, use:

        `symflow users list-identities`
    """
    table_data = get_user_data(options.sym_api)
    click.echo(tabulate(table_data, headers="firstrow"))


def get_user_data(api: SymAPI) -> List[List[str]]:
    table_data = [["Email", "Role", "Created At"]]
    for user in api.get_users():
        created_at = human_friendly(utc_to_local(user.created_at))  # type: ignore
        table_data.append([user.sym_email, "admin" if user.is_admin else "", created_at])

    return table_data



import csv
import tempfile
from typing import Dict

import click

from sym.flow.cli.helpers.api_operations import OperationHelper
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.users import UserUpdateSet
from sym.flow.cli.models.user import User


@click.command(name="update", short_help="Edit Users' Identities")
@click.option(
    "-i",
    "--input-file",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="File to use as input data",
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def users_update(options: GlobalOptions, input_file: str) -> None:
    """Create, delete or modify Sym Users and their identities such as AWS SSO IDs.
    If no input file is provided, this will trigger an interactive prompt to edit data in CSV format.

    \b
    Example:
        `symflow users update --input-file my-users.csv`
    """

    api = options.sym_api

    users = api.get_users()
    services = api.get_services()
    initial_data = UserUpdateSet(user_data=users, service_data=services)

    if input_file:
        filename = input_file
    else:
        click.confirm(
            "WARNING: This will open an editor in which you can modify ALL of your existing Users and Identities as raw CSV data!\nAre you sure you want to continue?",
            abort=True,
            default=False,
        )

        with tempfile.NamedTemporaryFile(mode="w", delete=False) as csv_file:
            writer = csv.DictWriter(
                csv_file, fieldnames=initial_data.headers, quotechar="`"
            )
            writer.writeheader()
            for user in initial_data.users:
                writer.writerow(prepare_user_row(initial_data=initial_data, user=user))

        click.edit(filename=csv_file.name, require_save=True)
        filename = csv_file.name

    edited_data = UserUpdateSet(service_data=services)
    with open(filename) as csv_update_file:
        reader = csv.DictReader(csv_update_file, quotechar="`")
        row_count = 2  # Start at 1 and skip header row
        try:
            for row in reader:
                if extra_services := (
                    {s for s in row.keys() if s} - set(initial_data.headers)
                ):
                    click.secho(
                        f"The following Services were provided but do not exist: {', '.join(extra_services)}",
                        fg="red",
                    )
                    return

                if (
                    None in row.keys()
                ):  # DictReader puts values without headers into a list under None
                    click.secho(
                        f"Warning: row {row_count} had unexpected formatting. Please verify your data to ensure it is correct.",
                        fg="yellow",
                    )

                edited_data.add_csv_row(row)
                row_count += 1

        except StopIteration:
            click.echo("No CSV data was provided. Exiting!")
            return

    operations = UserUpdateSet.compare_user_sets(initial_data, edited_data)
    operation_helper = OperationHelper(options, operations=operations)
    operation_helper.apply_changes()


def prepare_user_row(initial_data: UserUpdateSet, user: User) -> Dict[str, str]:
    row = {}
    for service in initial_data.services:
        if service.service_type in initial_data.uneditable_service_types:
            continue
        repr_identity = user.identity_repr_for_service(service)
        row[service.service_key] = repr_identity
    row["User ID"] = user.id
    return row



import click
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.commands.policies.generate import policies_generate
from sym.flow.cli.commands.policies.wizard import policies_wizard


@click.group(cls=SymGroup, short_help="Manipulate IAM Strategies")
def policies() -> None:
    pass


policies.add_command(policies_wizard)
policies.add_command(policies_generate)



from typing import List

from inquirer.questions import Checkbox

from sym.flow.cli.helpers.iam_policies import IAMPolicies
from sym.flow.cli.helpers.questions import ask


def format_actions(service: str, actions: List[str]) -> str:
    truncated = [action.split(":")[-1] for action in actions[:2]]
    return f"{service} ({', '.join(truncated)}, ...)"


def checkbox_for_access_levels(service: str) -> Checkbox:
    choices = [
        (format_actions(k, v), k)
        for k, v in IAMPolicies.actions_for_service(service).items()
        if v
    ]
    return Checkbox(
        name=service, message=f"Select {service} access levels", choices=choices
    )


def prompt_for_access_levels(service: str):
    return ask([checkbox_for_access_levels(service)])[service]



import json
import uuid
from datetime import timedelta
from typing import Iterator, List
from unittest.mock import ANY, patch
from uuid import UUID

import pytest
import requests_mock
from pydantic import parse_obj_as

from sym.flow.cli.errors import (
    NotAuthorizedError,
    SymAPIAggregateError,
    SymAPIError,
    SymAPIUnknownError,
)
from sym.flow.cli.helpers.api import SymAPI, SymRESTClient
from sym.flow.cli.models.resource import ResourceType, TerraformResource
from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import ServiceType
from sym.flow.cli.models.user import User
from sym.flow.cli.tests.conftest import get_mock_response
from sym.flow.cli.tests.factories.tokens import SymTokenFactory
from sym.flow.cli.tests.factories.users import UserFactory

MOCK_INTEGRATIONS_BAD_DATA = [
    {
        "name": "integration 1",
        "type": "aws",
    },
    {
        "name": "integration 2",
        "type": "aws_sso",
    },
]

MOCK_INTEGRATIONS_DATA = [
    {
        "id": "de6bd4ae-b30e-4033-9b06-bc918db2e619",
        "slug": "integration-1",
        "type": "aws",
        "srn": "test_org:sym_integration:aws:integration-1",
        "updated_at": "2021-01-19 19:29:46.505678+00",
    },
    {
        "id": "e8e46254-585d-4cac-9744-fd88a3c81390",
        "slug": "integration-2",
        "srn": "test_org:sym_integration:aws_sso:integration-2",
        "type": "aws_sso",
        "updated_at": "2021-01-19 18:29:46.505678+00",
    },
]

MOCK_SERVICES_DATA = {
    "services": [
        {
            "id": "e016f857-9979-4710-a53b-0dd6cedcae53",
            "slug": "google",
            "label": "Google",
            "external_id": "symops.io",
        },
        {
            "id": "166e7fef-c44b-40ce-b6ed-d3066799060a",
            "slug": "sym",
            "label": "Sym",
            "external_id": "cloud",
        },
        # This service will be filtered out by the get_services api because the service type is not recognized
        {
            "id": "abc",
            "slug": "cloud",
            "label": "Some service not in Service Types",
            "external_id": "cloud",
        },
    ]
}

MOCK_CREATE_SERVICE_DATA = {
    "status": "success",
    "errors": [],
    "succeeded": 1,
    "failed": 0,
    "results": [
        {
            "id": "4df1dae0-892b-47a1-9dee-9f818b7d331b",
            "slug": "slack",
            "label": "Slack",
            "external_id": "T123ABC",
        }
    ],
}

MOCK_SERVICE_REFERENCES_DATA = {
    "error": "false",
    "references": {
        "identities": ["274ab5cf-7322-430a-97eb-e7f3cec6a987"],
        "integrations": [],
    },
}


MOCK_BASE_URL = "http://faketest.symops.io/api/v1"


@pytest.fixture
def sym_rest_client(sandbox) -> Iterator[SymRESTClient]:
    with sandbox.push_xdg_config_home():
        yield SymRESTClient(url=MOCK_BASE_URL, access_token="")


@pytest.fixture
def sym_api(sandbox) -> Iterator[SymAPI]:
    with sandbox.push_xdg_config_home():
        yield SymAPI(url=MOCK_BASE_URL, access_token="abc")


class TestSymRESTClient:
    def test_make_headers(self, sym_rest_client, auth_token):
        with pytest.raises(NotAuthorizedError, match="symflow login"):
            sym_rest_client.make_headers()

        sym_rest_client.access_token = "access"
        headers = sym_rest_client.make_headers()
        assert headers.get("X-Sym-Request-ID") is not None
        assert headers.get("Authorization") == "Bearer access"

    def test_handle_response(self, sym_rest_client):
        response_500 = get_mock_response(500)
        with pytest.raises(SymAPIUnknownError, match="500"):
            sym_rest_client.handle_response(response_500)

        response_400 = get_mock_response(400)
        with pytest.raises(SymAPIUnknownError, match="400"):
            sym_rest_client.handle_response(response_400)

        response_with_msg = get_mock_response(
            401,
            {
                "error": True,
                "message": "The provided JWT could not be validated.",
                "code": "AuthenticationError:INVALID_JWT",
                "status_code": 401,
                "is_retryable": False,
            },
        )
        with pytest.raises(SymAPIAggregateError, match=response_with_msg.data["message"]):
            sym_rest_client.handle_response(response_with_msg)

        response_with_error_list = get_mock_response(
            400,
            {
                "status": "error",
                "errors": ["some error message"],
                "succeeded": 0,
                "failed": 1,
                "results": [],
            },
        )
        with pytest.raises(
            SymAPIAggregateError, match=response_with_error_list.data["errors"][0]
        ):
            sym_rest_client.handle_response(response_with_error_list)

        response_200 = get_mock_response(200)
        assert sym_rest_client.handle_response(response_200) == response_200

    @patch(
        "requests.request",
        return_value=get_mock_response(200),
    )
    def test_request(self, mock_request, sym_rest_client):
        data = {"test": "hello"}

        # Auth is required by default, make sure if force_auth not specified
        with pytest.raises(NotAuthorizedError, match="symflow login"):
            sym_rest_client.post("fake-endpoint", data)

        mock_request.assert_not_called()

        # If force_auth specified False, we should let the call through
        sym_rest_client._request("POST", "fake-endpoint", force_auth=False, data=data)
        mock_request.assert_called_with(
            "POST",
            url=f"{MOCK_BASE_URL}/fake-endpoint",
            data=json.dumps(data),
            headers=ANY,
            params=ANY,
        )

        # If force_auth not specified False but access token exists, let call through
        sym_rest_client.access_token = "access"
        sym_rest_client.post("fake-endpoint", data)
        mock_request.assert_called_with(
            "POST",
            url=f"{MOCK_BASE_URL}/fake-endpoint",
            data=json.dumps(data),
            headers=ANY,
            params=ANY,
        )


class TestSymAPI:
    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200, {"client_id": "12345abc", "slug": "test"})),
    )
    def test_get_organization_from_slug(self, mock_api_get, sym_api, test_org):
        org_data = sym_api.get_organization_from_slug("test")

        mock_api_get.assert_called_once_with("auth/org/test", force_auth=False)

        assert org_data.slug == test_org.slug
        assert org_data.client_id == test_org.client_id

    @pytest.mark.parametrize("status_code", [400, 403, 500])
    def test_verify_login_failure(self, status_code, sym_api):
        with patch(
            "sym.flow.cli.helpers.api.SymRESTClient.get",
            return_value=(get_mock_response(status_code)),
        ) as mock_api_get:
            assert sym_api.verify_login("test@symops.io") is False

        mock_api_get.assert_called_once_with(
            "auth/login", {"email": "test@symops.io"}, validate=False
        )

    @pytest.mark.parametrize("status_code", [200])
    def test_verify_login_success(self, status_code, sym_api):
        with patch(
            "sym.flow.cli.helpers.api.SymRESTClient.get",
            return_value=(get_mock_response(status_code)),
        ) as mock_api_get:
            assert sym_api.verify_login("test@symops.io") is True
        mock_api_get.assert_called_once_with(
            "auth/login", {"email": "test@symops.io"}, validate=False
        )

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200, data=MOCK_INTEGRATIONS_DATA)),
    )
    def test_get_integrations(self, _, sym_api):
        data = sym_api.get_integrations()
        assert len(data) == 2
        assert data[0]["slug"] == "integration-1"
        assert data[0]["type"] == "aws"
        assert data[0]["srn"] == "test_org:sym_integration:aws:integration-1"
        assert data[0]["updated_at"] == "2021-01-19 19:29:46.505678+00"
        assert data[1]["slug"] == "integration-2"
        assert data[1]["type"] == "aws_sso"
        assert data[1]["srn"] == "test_org:sym_integration:aws_sso:integration-2"
        assert data[1]["updated_at"] == "2021-01-19 18:29:46.505678+00"

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200, data=MOCK_INTEGRATIONS_DATA)),
    )
    def test_get_resources(self, _, sym_api):
        data = sym_api.get_resources(resource_type=ResourceType.SYM_INTEGRATION)
        assert len(data) == 2

        resource = data[0]
        assert isinstance(resource, TerraformResource)
        assert resource.identifier == UUID("de6bd4ae-b30e-4033-9b06-bc918db2e619")
        assert resource.sub_type == "aws"
        assert resource.slug == "integration-1"

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200, data=MOCK_SERVICES_DATA)),
    )
    def test_get_services(self, mock_api_get, sym_api):
        all_services = parse_obj_as(List[Service], MOCK_SERVICES_DATA["services"])
        expected_services = [
            s for s in all_services if s.service_type in ServiceType.all_names()
        ]

        data = sym_api.get_services()
        mock_api_get.assert_called_once_with("services")
        assert data == expected_services

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(
            get_mock_response(200, {"services": MOCK_SERVICES_DATA["services"][:1]})
        ),
    )
    def test_get_service(self, mock_api_get, sym_api):
        expected = Service.parse_obj(MOCK_SERVICES_DATA["services"][0])

        data = sym_api.get_service(expected.service_type, expected.external_id)
        mock_api_get.assert_called_once_with(
            "services",
            {
                "service_type": expected.service_type,
                "external_id": expected.external_id,
            },
        )
        assert data == expected

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200, MOCK_SERVICES_DATA)),
    )
    def test_get_service_unexpected_response(self, mock_api_get, sym_api):
        # Mock returns more than one service
        with pytest.raises(SymAPIError, match="Expected 1 service, but found 3 instead"):
            sym_api.get_service("slack", "T123")

        mock_api_get.assert_called_once_with(
            "services", {"service_type": "slack", "external_id": "T123"}
        )

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.post",
        return_value=(get_mock_response(200, data=MOCK_CREATE_SERVICE_DATA)),
    )
    def test_create_service(self, mock_api_post, sym_api):
        service_type = ServiceType.SLACK
        external_id = "T123ABC"
        service = sym_api.create_service(service_type, external_id)
        mock_api_post.assert_called_once_with(
            "services",
            {"service_type": service_type.type_name, "external_id": external_id},
        )
        assert service.service_type == service_type.type_name
        assert service.external_id == external_id

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.post",
        return_value=(get_mock_response(200)),
    )
    def test_delete_services(self, mock_api_post, sym_api):
        service_type = ServiceType.SLACK
        external_id = "T123ABC"
        sym_api.delete_service(service_type.type_name, external_id)
        mock_api_post.assert_called_once_with(
            "services/delete",
            {"service_type": service_type.type_name, "external_id": external_id},
        )

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200, MOCK_SERVICE_REFERENCES_DATA)),
    )
    def test_get_service_references(self, mock_api_get, sym_api):

        data = sym_api.get_service_references("fake-service-id")
        mock_api_get.assert_called_once_with(
            "service/fake-service-id/references",
        )
        assert data == MOCK_SERVICE_REFERENCES_DATA["references"]

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200, {"url": "http://hello.sym.com"})),
    )
    def test_get_slack_install_url(self, mock_api_get, sym_api):
        url = sym_api.get_slack_install_url("fake_service_id")
        mock_api_get.assert_called_once_with(
            "services/slack/link", {"service_id": "fake_service_id"}
        )
        assert url == "http://hello.sym.com"

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200)),
    )
    def test_uninstall_slack(self, mock_api_get, sym_api):
        assert sym_api.uninstall_slack(service_id="T1234567") is None
        mock_api_get.assert_called_once_with(
            "services/slack/uninstall", {"token": "T1234567"}, validate=False
        )

    def test_last_request_id(self, sym_api):
        with requests_mock.Mocker() as m:
            m.get(
                f"{MOCK_BASE_URL}/auth/org/fake-org",
                text='{"slug": "fake-org", "client_id": "fake-client-id"}',
            )

            assert not sym_api.rest.last_request_id
            resp = sym_api.get_organization_from_slug("fake-org")
            assert resp.slug == "fake-org"
            assert sym_api.rest.last_request_id

    def test_get_users(self, sym_api: SymAPI):
        users: List[User] = UserFactory.create_batch(5)
        with patch(
            "sym.flow.cli.helpers.api.SymRESTClient.get",
            return_value=(get_mock_response(200, {"users": users})),
        ) as mock_api_get:
            data = sym_api.get_users()
            mock_api_get.assert_called_once_with("users", None)
            assert data == users

    def test_get_user(self, sym_api: SymAPI):
        users: List[User] = UserFactory.create_batch(5)
        with patch(
            "sym.flow.cli.helpers.api.SymRESTClient.get",
            return_value=(get_mock_response(200, {"users": users})),
        ) as mock_api_get:
            data = sym_api.get_user(users[0].sym_email)
            mock_api_get.assert_called_once_with("users", None)
            assert data == users[0]

    def test_get_user_by_id(self, sym_api: SymAPI):
        users: List[User] = UserFactory.create_batch(5)
        with patch(
            "sym.flow.cli.helpers.api.SymRESTClient.get",
            return_value=(get_mock_response(200, {"users": users[0:1]})),
        ) as mock_api_get:
            user_uuid = uuid.UUID(users[0].id)
            data = sym_api.get_user_by_id(user_uuid)
            mock_api_get.assert_called_once_with("users", {"id": user_uuid})
            assert data == users[0]

    def test_create_token(self, sym_api: SymAPI):
        access_token = str(uuid.uuid4())
        with patch(
            "sym.flow.cli.helpers.api.SymRESTClient.post",
            return_value=(get_mock_response(200, {"access_token": access_token})),
        ) as mock_api:
            username = "bot-user"
            expiry = int(timedelta(seconds=30).total_seconds())
            label = "some label"
            data = sym_api.create_token(username, expiry, label)
            mock_api.assert_called_once_with(
                "tokens",
                {"username": username, "expiry": expiry, "label": label},
            )
            assert data == access_token

    def test_create_token_no_label(self, sym_api: SymAPI):
        access_token = str(uuid.uuid4())
        with patch(
            "sym.flow.cli.helpers.api.SymRESTClient.post",
            return_value=(get_mock_response(200, {"access_token": access_token})),
        ) as mock_api:
            username = "bot-user"
            expiry = int(timedelta(seconds=30).total_seconds())
            data = sym_api.create_token(username, expiry)
            mock_api.assert_called_once_with(
                "tokens", {"username": username, "expiry": expiry}
            )
            assert data == access_token

    def test_revoke_token(self, sym_api: SymAPI):
        with patch(
            "sym.flow.cli.helpers.api.SymRESTClient.post",
            return_value=(get_mock_response(200)),
        ) as mock_api:
            identifier = str(uuid.uuid4())
            sym_api.revoke_token(identifier)
            mock_api.assert_called_once_with("tokens/delete", {"identifier": identifier})

    def test_get_tokens(self, sym_api: SymAPI):
        tokens = SymTokenFactory.create_batch(3)
        with patch(
            "sym.flow.cli.helpers.api.SymRESTClient.get",
            return_value=(get_mock_response(200, {"tokens": tokens})),
        ) as mock_api:
            result = sym_api.get_tokens()

            mock_api.assert_called_once_with("tokens")
            assert result == tokens



from enum import Enum
from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel, Field


# Enumeration of all valid Terraform resource types, with values matching
# the corresponding Sym API entity name.
class ResourceType(str, Enum):
    SYM_FLOW = "flows"
    SYM_TARGET = "access-targets"
    SYM_INTEGRATION = "integrations"
    SYM_SECRET = "secrets"
    SYM_SECRETS = "secret-sources"
    SYM_STRATEGY = "access-strategies"
    SYM_LOG_DESTINATION = "log-destinations"
    SYM_ENVIRONMENT = "environments"
    SYM_RUNTIME = "runtimes"
    SYM_ERROR_LOGGER = "error-loggers"

    @classmethod
    def options(cls) -> List[str]:
        """Returns a list of valid resource types represented as their Terraform names."""
        return [tf_name.lower() for tf_name, _ in cls.__members__.items()]


class TerraformResource(BaseModel):
    """A model for parsing generic Terraform resources from the API."""

    identifier: UUID = Field(alias="id")
    slug: str
    srn: str
    sub_type: Optional[str] = Field(
        alias="type", default=None
    )  # e.g. "slack" for integrations



from typing import Optional

import click

from sym.flow.cli.commands.services.hooks.slack_delete import slack_delete
from sym.flow.cli.commands.users.utils import get_or_prompt_service
from sym.flow.cli.errors import ReferencedObjectError
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.utils import filter_dict
from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import ServiceType


def pre_service_delete_hooks(options: GlobalOptions, service: Service) -> None:
    """Registered hooks to call after before deletion of services"""

    if service.service_type == ServiceType.SLACK.type_name:
        slack_delete(options.sym_api, service.id)


@click.command(name="delete", short_help="Delete an existing service")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.option(
    "--service-type",
    "_service_type",
    help="The service to delete",
)
@click.option(
    "--external-id",
    "_external_id",
    help="The identifier for the service",
)
def services_delete(
    options: GlobalOptions, _service_type: Optional[str], _external_id: Optional[str]
) -> None:
    """Set up a new service for your organization."""
    service = get_or_prompt_service(options.sym_api, _service_type, _external_id)

    active_references = filter_dict(
        options.sym_api.get_service_references(service.id), lambda refs: len(refs) > 0
    )
    if active_references:
        raise ReferencedObjectError(active_references)

    pre_service_delete_hooks(options, service)
    options.sym_api.delete_service(service.service_type, service.external_id)
    click.secho(
        f"Successfully deleted service type {service.service_type} with external ID {service.external_id}!",
        fg="green",
    )



import uuid
from unittest.mock import MagicMock, patch

import pytest

from sym.flow.cli.commands.services.hooks.slack_new import get_magic_url, slack_new
from sym.flow.cli.errors import NotAuthorizedError
from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.constants import DEFAULT_API_URL
from sym.flow.cli.tests.conftest import get_mock_response


class TestSlackNewCreateHook:
    """Suite for testing Slack installation."""

    MOCK_MAGIC_URL = "http://fake-test.symops.com"

    @patch("sym.flow.cli.commands.services.hooks.slack_new.webbrowser.open")
    @patch(
        "sym.flow.cli.commands.services.hooks.slack_new.click.confirm",
        return_value=True,
    )
    @patch(
        "sym.flow.cli.commands.services.hooks.slack_new.get_magic_url",
        return_value=MOCK_MAGIC_URL,
    )
    def test_slack_install_open_browser(
        self, mock_get_magic_url, mock_click_confirm, mock_webbrowser_open
    ):
        fake_service_id = str(uuid.uuid4())
        api = SymAPI(DEFAULT_API_URL)
        slack_new(api, fake_service_id)
        mock_get_magic_url.assert_called_once_with(api=api, service_id=fake_service_id)
        mock_click_confirm.assert_called_once()
        mock_webbrowser_open.assert_called_once_with(self.MOCK_MAGIC_URL)

    @patch("sym.flow.cli.commands.services.hooks.slack_new.webbrowser.open")
    @patch(
        "sym.flow.cli.commands.services.hooks.slack_new.click.confirm",
        return_value=False,
    )
    @patch("sym.flow.cli.commands.services.hooks.slack_new.click.secho")
    @patch(
        "sym.flow.cli.commands.services.hooks.slack_new.get_magic_url",
        return_value=MOCK_MAGIC_URL,
    )
    def test_click_command_no_browser(
        self,
        mock_get_magic_url,
        mock_click_secho,
        mock_click_confirm,
        mock_webbrowser_open,
    ):
        fake_service_id = str(uuid.uuid4())
        api = SymAPI(DEFAULT_API_URL)
        slack_new(api, fake_service_id)

        assert self.MOCK_MAGIC_URL in mock_click_secho.call_args.args[0]

        mock_get_magic_url.assert_called_once_with(api=api, service_id=fake_service_id)
        mock_click_confirm.assert_called_once()
        mock_webbrowser_open.assert_not_called()

    @patch(
        "sym.flow.cli.commands.services.hooks.slack_new.get_magic_url",
        side_effect=ValueError("random error"),
    )
    def test_click_call_catches_unknown_error(self, mock_get_magic_url, click_setup):
        with pytest.raises(ValueError, match="random error"):
            fake_service_id = str(uuid.uuid4())
            slack_new(MagicMock(), fake_service_id)

            mock_get_magic_url.assert_called_once()

    def test_initialize_slack_install_not_authorized_fails(self, sandbox, global_options):
        with pytest.raises(NotAuthorizedError, match="symflow login"):
            with sandbox.push_xdg_config_home():
                get_magic_url(global_options.sym_api, "uuid")

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200, data={"url": "http://test.symops.io"})),
    )
    def test_get_magic_url(self, mock_api_get, sandbox, global_options):
        with sandbox.push_xdg_config_home():
            assert (
                get_magic_url(global_options.sym_api, str(uuid.uuid4()))
                == "http://test.symops.io"
            )

        mock_api_get.assert_called_once()



import click
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.helpers.global_options import GlobalOptions

from .add import domains_add
from .list import domains_list
from .remove import domains_remove


@click.group(name="domains", cls=SymGroup, short_help="Perform operations on domains")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def domains(options: GlobalOptions) -> None:
    """Operations on domains.

    Domains are used to decide who can join your organization. For example,
    when an unknown user interacts with your Sym Slack App, they will be automatically
    added to your organization (and allowed to run your Flows) if and only if their email
    domain matches one defined for your organization.

    Users may still be added manually using `symflow users create` even if their email does not
    match any of the configured domains.
    """


domains.add_command(domains_add)
domains.add_command(domains_list)
domains.add_command(domains_remove)



import os

import click
from sym.shared.cli.decorators import setup_segment, setup_sentry
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.commands.bots import bots_commands
from sym.flow.cli.commands.config import config_commands
from sym.flow.cli.commands.debug import debug_commands
from sym.flow.cli.commands.domains import domains_commands
from sym.flow.cli.commands.login import login, logout
from sym.flow.cli.commands.policies import policies_commands
from sym.flow.cli.commands.resources import resources_commands
from sym.flow.cli.commands.services import services_commands
from sym.flow.cli.commands.status import status
from sym.flow.cli.commands.tokens import tokens_commands
from sym.flow.cli.commands.users import users_commands
from sym.flow.cli.commands.version import version
from sym.flow.cli.helpers.constants import (
    DEFAULT_API_URL,
    DEFAULT_AUTH_URL,
    SegmentWriteKey,
    SentryDSN,
)
from sym.flow.cli.helpers.version import maybe_display_update_message

from ..helpers.global_options import GlobalOptions
from ..version import __version__


@click.group(
    name="symflow", cls=SymGroup, context_settings={"help_option_names": ["-h", "--help"]}
)
@click.option(
    "--debug", is_flag=True, help="enable verbose debugging", envvar="SYM_DEBUG"
)
@click.option(
    "--api-url",
    default=DEFAULT_API_URL,
    help="set the Sym API URL",
    envvar="SYM_API_URL",
)
@click.option(
    "--auth-url",
    default=DEFAULT_AUTH_URL,
    help="set the Sym auth url",
    envvar="SYM_AUTH_URL",
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
@setup_segment(write_key=SegmentWriteKey)
@setup_sentry(dsn=SentryDSN, release=f"sym-flow-cli@{__version__}")
def symflow(options: GlobalOptions, auth_url: str, api_url: str, debug: bool) -> None:
    """
    Manage Sym Services and Users in your organization

    https://docs.symops.com/docs/using-the-symflow-cli
    """

    options.debug = debug
    if sym_jwt := os.getenv("SYM_JWT"):
        options.set_access_token(sym_jwt)
    options.set_api_url(api_url)
    options.set_auth_url(auth_url)
    options.dprint(auth_url=auth_url, api_url=api_url)

    maybe_display_update_message()


symflow.add_command(version)
symflow.add_command(login)
symflow.add_command(logout)
symflow.add_command(status)
symflow.add_command(services_commands)
symflow.add_command(users_commands)
symflow.add_command(debug_commands)
symflow.add_command(config_commands)
symflow.add_command(policies_commands)
symflow.add_command(tokens_commands)
symflow.add_command(bots_commands)
symflow.add_command(resources_commands)
symflow.add_command(domains_commands)



from typing import List

import click

from sym.flow.cli.models.service_type import ServiceType


class ServiceTypeChoice(click.Choice):
    """Custom click.Choice for ServiceType Enums

    Sets the choice to the given list of services (defaults to all), then converts the input str value to the ServiceType Enum
    """

    name = "service_type_choice"

    def __init__(self, choices: List[ServiceType] = None):
        service_choices = choices or ServiceType.public()

        super().__init__(
            choices=[s.value.type_name for s in service_choices], case_sensitive=False
        )

    def convert(self, value, param, ctx) -> ServiceType:
        if isinstance(value, str):
            # Values entered directly as --service-type are passed through as str from click
            service_type_str = super().convert(value, param, ctx)
            # Service Types are defined as all uppercase enums
            return ServiceType[service_type_str.upper()]
        else:
            # Values selected by inquirer are already the correct type
            return value



import click


class ExternalIdOption(click.Option):
    """Custom click.Option whose prompt dynamically changes depending on the "service_type" value inputted"""

    def __init__(self, *args, **kwargs):
        super(ExternalIdOption, self).__init__(*args, **kwargs)

    def prompt_for_value(self, ctx):
        """Dynamically change external_id prompt text based on service_type"""
        # ServiceTypes Enums are defined in all caps
        self.prompt = ctx.params["service_type"].external_id_name

        return super().prompt_for_value(ctx)



from typing import Optional

import click

import sym.flow.cli.helpers.cli as cli_helpers
from sym.flow.cli.errors import SymAPIRequestError
from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.global_options import GlobalOptions


@click.command(short_help="Check your stored auth token")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def status(options: GlobalOptions) -> None:
    """Check if you have an existing login session that is still valid."""
    if options.access_token:
        check_status_by_env(options.sym_api)
    else:
        check_status_by_config(options.sym_api)


def check_status_by_env(api: SymAPI):
    validate_login(api)
    click.echo(f"   You are logged in via SYM_JWT.")


def check_status_by_config(api: SymAPI):
    if not Config.is_logged_in():
        cli_helpers.fail("You are not currently logged in", "Try running `symflow login`")

    org = Config.get_org().slug
    email = Config.get_email()
    validate_login(api, email)

    click.echo(
        f"   You are logged in to {click.style(org, bold=True)} as {click.style(email, bold=True)}."
    )


def validate_login(api: SymAPI, email: Optional[str] = None):
    try:
        if not api.verify_login(email):
            cli_helpers.fail("Token expired", "Try running `symflow login`")

        click.secho("✔️  Status check succeeded!", fg="green")

    except SymAPIRequestError:
        cli_helpers.fail("A server error has occurred. Please try again later")



import click

from sym.flow.cli.helpers.global_options import GlobalOptions


@click.command(name="add", short_help="Add a domain")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.argument("domain")
def domains_add(options: GlobalOptions, domain: str) -> None:
    """Add a new domain to your organization."""

    options.sym_api.add_domain(domain)
    click.echo(f"{domain} successfully added as a domain.")



from datetime import datetime
from typing import Optional
from uuid import UUID

from pydantic import BaseModel


class SymAccessToken(BaseModel):
    """Pydantic model for parsing Sym response after creating a Sym Token.
    The output should be a base64 encoded JWT
    """

    access_token: str


class SymTokenUser(BaseModel):
    """Pydantic model for parsing the simplified UserSchema outputted by GET /api/v1/tokens"""

    id: UUID
    type: str


class SymToken(BaseModel):
    """Pydantic model for parsing Sym response to GET /api/v1/tokens"""

    identifier: UUID
    user: SymTokenUser
    created_by: SymTokenUser
    expires_at: datetime
    updated_at: datetime
    label: Optional[str]



import click

from sym.flow.cli.helpers.api import SymAPI


def slack_delete(api: SymAPI, service_id: str) -> None:
    """Pre-delete service hook to uninstall the Sym Slack App from a workspace.
    This will completely revoke Sym's access in the workspace.

    To see what workspaces the Sym App is installed in, and the corresponding
    Workspace IDs, use `symflow services list`.
    """

    api.uninstall_slack(service_id)
    click.echo(
        "Uninstall successful! The Sym App has been removed from your Slack workspace."
    )



from enum import Enum
from typing import Callable, List, NamedTuple, Optional


class ServiceTypeMeta(NamedTuple):
    type_name: str
    matcher: str
    external_id_name: str
    validator: Optional[Callable[[str], bool]] = None
    help_str: Optional[str] = None


class ServiceType(ServiceTypeMeta, Enum):
    """All Service Types recognized by the Sym API"""

    @classmethod
    def public(cls) -> List["ServiceType"]:
        """Returns all ServiceTypes that can be edited by external users"""
        private_services = [cls.SYM, cls.AUTH0]
        return [s for s in cls if s not in private_services]

    @classmethod
    def all_public_names(cls) -> List[str]:
        """Returns all service types names that can be edited by external users"""
        return [str(s) for s in cls.public()]

    @classmethod
    def all_names(cls) -> List[str]:
        """Returns all service type names"""
        return [str(s) for s in ServiceType]

    @classmethod
    def get(cls, name) -> Optional["ServiceType"]:
        # Names in the enum member map are all uppercase
        if name.upper() in cls._member_map_:
            return ServiceType[name.upper()]
        else:
            return None

    def __str__(self):
        return self.value.type_name

    SYM = ServiceTypeMeta("sym", matcher="email", external_id_name="Org Slug")
    SLACK = ServiceTypeMeta(
        "slack",
        matcher="user_id",
        external_id_name="Workspace ID",
        validator=lambda id: id.startswith("T") and id.isupper(),
        help_str="Must start with 'T' and be all upper case",
    )
    AWS_SSO = ServiceTypeMeta(
        "aws_sso",
        matcher="principal_uuid",
        external_id_name="Instance ARN",
        validator=lambda arn: arn.startswith("arn:aws:sso:::instance/"),
        help_str="Must start with 'arn:aws:sso:::instance/'",
    )
    PAGERDUTY = ServiceTypeMeta(
        "pagerduty", matcher="user_id", external_id_name="PagerDuty Subdomain"
    )
    AWS_IAM = ServiceTypeMeta(
        "aws_iam",
        matcher="user_arn",
        external_id_name="Account #",
        validator=lambda acct_num: acct_num.isdigit() and len(acct_num) == 12,
        help_str="Must be a 12 digit account number",
    )
    APTIBLE = ServiceTypeMeta(
        "aptible", matcher="user_id", external_id_name="Organization ID"
    )
    GOOGLE = ServiceTypeMeta("google", matcher="email", external_id_name="Email Domain")
    AUTH0 = ServiceTypeMeta("auth0", matcher="idp_id", external_id_name="External ID")
    OKTA = ServiceTypeMeta("okta", matcher="user_id", external_id_name="Okta Domain")
    GITHUB = ServiceTypeMeta(
        "github", matcher="user_id", external_id_name="Github Organization Name"
    )
    CUSTOM = ServiceTypeMeta(
        "custom", matcher="user_id", external_id_name="Custom Service External ID"
    )
    TAILSCALE = ServiceTypeMeta(
        "tailscale", matcher="email", external_id_name="Tailnet Name"
    )


# Services managed by Sym, not customers
MANAGED_SERVICES = [ServiceType.SYM.type_name, ServiceType.AUTH0.type_name]



import uuid
from contextlib import ExitStack
from unittest.mock import patch

import pytest

from sym.flow.cli.models import Organization
from sym.flow.cli.symflow import symflow as click_command


class TestDomains:
    @pytest.fixture(autouse=True)
    def mocks(self):
        mock_org = Organization(
            id=str(uuid.uuid4()), slug="test", domains=["test.invalid", "blah.symops.com"]
        )

        with ExitStack() as stack:
            stack.enter_context(
                patch(
                    "sym.flow.cli.helpers.api.SymAPI.get_current_organization",
                    return_value=mock_org,
                )
            )
            yield

    @patch("sym.flow.cli.helpers.rest.SymRESTClient.post")
    def test_domains_add(self, _, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["domains", "add", "test.invalid"],
            )
            assert result.exit_code == 0
            assert result.output == "test.invalid successfully added as a domain.\n"

    @patch("sym.flow.cli.helpers.rest.SymRESTClient.post")
    def test_domains_remove(self, _, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["domains", "remove", "test.invalid"],
            )
            assert result.exit_code == 0
            assert result.output == "test.invalid successfully removed as a domain.\n"

    def test_domains_list(self, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["domains", "list"],
            )
            assert result.exit_code == 0
            assert "- test.invalid\n- blah.symops.com\n" in result.output



from unittest.mock import call, patch

import pytest

from sym.flow.cli.commands.users.create import prompt_for_identity
from sym.flow.cli.models.service_type import ServiceType
from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.factories.users import ServiceFactory, UserFactory


@patch("sym.flow.cli.helpers.api_operations.OperationHelper.handle_update_users")
class TestUsersCreate:
    @pytest.fixture
    def users(self):
        yield UserFactory.create_batch(3)

    @pytest.fixture
    def services(self, users):
        # Create a second slack service
        slack2 = ServiceFactory.create(slug=ServiceType.SLACK.type_name)
        all_services = [
            i.service for i in users[0].identities if i.service.service_type
        ] + [slack2]

        # User Factory creates the same services for each user
        yield all_services

    @pytest.fixture(autouse=True)
    def patchypatch(self, users, services):
        with patch("sym.flow.cli.helpers.api.SymAPI.get_users", return_value=users):
            with patch(
                "sym.flow.cli.helpers.api.SymAPI.get_services", return_value=services
            ):
                yield

    def test_create_blank_user(self, mock_apply, click_setup, users):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "users",
                    "create",
                    "foo_bar@symops.io",
                ],
            )
            assert result.exit_code == 0

    def test_create_user_existing_email(self, mock_apply, click_setup, users):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["users", "create", users[0].sym_email],
            )
            assert (
                f"A user already exists with email: {users[0].sym_email}" in result.output
            )

    def test_create_user_prompt_identity(self, mock_apply, click_setup, users, services):
        with click_setup() as runner:
            with patch(
                "inquirer.text",
                side_effect=["U123", "U456"],  # two slack services, should inquire twice
            ) as mock_inquire:
                result = runner.invoke(
                    click_command,
                    [
                        "users",
                        "create",
                        "foo_bar@symops.io",
                        "-s",
                        "slack",
                    ],
                )
                assert result.exit_code == 0
                expected_calls = [
                    call(f"Slack ({s.external_id}) {ServiceType.SLACK.matcher}")
                    for s in services
                    if s.service_type == ServiceType.SLACK.type_name
                ]
                mock_inquire.assert_has_calls(expected_calls, any_order=True)

    def test_create_user_unknown_service(
        self,
        mock_apply,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "users",
                    "create",
                    "foo_bar@symops.io",
                    "-s",
                    "something-bad",
                ],
            )
            assert "Invalid input: 'something-bad'" in result.output
            mock_apply.assert_not_called()

    def test_create_user_not_registered_service(
        self,
        mock_apply,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "users",
                    "create",
                    "foo_bar@symops.io",
                    "-s",
                    "google",  # factory does not create an google service
                ],
            )
            assert "No service is registered for type google" in result.output
            assert (
                "You can create the service with `symflow services create --service-type google`"
                in result.output
            )
            mock_apply.assert_not_called()

    @patch("inquirer.text", return_value="  ")
    def test_prompt_for_identity_skips_blank_input(self, mock_inquirer, mock_apply):
        service = ServiceFactory.create(slug=ServiceType.SLACK.type_name)
        result = prompt_for_identity(ServiceType.get(service.service_type), service)  # type: ignore
        expected_calls = [
            call(f"Slack ({service.external_id}) {ServiceType.SLACK.matcher}")
        ]
        mock_inquirer.assert_has_calls(expected_calls)

        assert result is None



from contextlib import ExitStack, contextmanager
from pathlib import Path
from typing import Optional, Union
from unittest.mock import patch

import pytest
from requests import Response
from sym.shared.cli.helpers.config import init
from sym.shared.cli.helpers.contexts import push_env
from sym.shared.cli.tests.conftest import capture_command, wrapped_cli_runner  # noqa

from sym.flow.cli.helpers.constants import DEFAULT_API_URL
from sym.flow.cli.helpers.jwt import JWT
from sym.flow.cli.models import AuthToken, Organization
from sym.flow.cli.tests.helpers.sandbox import Sandbox


@pytest.fixture
def click_setup(sandbox, wrapped_cli_runner):
    @contextmanager
    def context():
        with wrapped_cli_runner.isolated_filesystem():
            with sandbox.push_xdg_config_home():
                with sandbox.push_exec_path():
                    with push_env("SYM_API_URL", DEFAULT_API_URL):
                        yield wrapped_cli_runner

    return context


@pytest.fixture
def sandbox(tmp_path: Path) -> Sandbox:
    return Sandbox(tmp_path)


@pytest.fixture
def test_org() -> Organization:
    return Organization(slug="test", client_id="12345abc")


@pytest.fixture
def auth_token():
    return AuthToken(
        access_token="access",
        refresh_token="refresh",
        token_type="type",
        expires_in=86400,
        scope="scopes",
    )


@pytest.fixture(autouse=True)
def init_cli():
    init("symflow")


@pytest.fixture
def global_options():
    from sym.flow.cli.helpers.global_options import GlobalOptions

    options = GlobalOptions()
    options.debug = False
    options.api_url = "https://api.com"
    options.auth_url = "https://auth.com"
    return options


@pytest.fixture
def mock_jwt():
    return JWT(
        client_id="12345",  # type: ignore
        connection_strategy="google-oauth2",  # type: ignore
        connection_name="google-oauth2",  # type: ignore
        email="test@symops.io",  # type: ignore
        name="Test User",  # type: ignore
    )


@pytest.fixture(autouse=True)
def global_mocks(request, mock_jwt):
    if "skip_global_mocks" not in request.keywords:
        with ExitStack() as stack:
            stack.enter_context(
                patch(
                    "sym.flow.cli.helpers.version.get_latest_version", return_value=None
                )
            )
            stack.enter_context(
                patch(
                    "sym.flow.cli.helpers.jwt.JWT.from_access_token",
                    return_value=mock_jwt,
                )
            )
            stack.enter_context(patch("webbrowser.open"))
            yield


class MockResponse(Response):
    def __init__(self, data: Optional[Union[dict, list]] = None):
        super().__init__()

        if data is None:
            data = {}

        self.data = data

    def json(self):
        return self.data


def get_mock_response(
    status_code: int, data: Optional[Union[dict, list]] = None, url: Optional[str] = None
) -> Response:
    response = MockResponse(data)
    response.status_code = status_code

    if url:
        response.url = url

    return response



from unittest.mock import patch
from uuid import UUID

from sym.shared.cli.tests.helpers.capture import CaptureCommand

from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.helpers.sandbox import Sandbox

JSON = '{"foo": "bar"}'
UUID_VAL = "6fac5fa7-898f-4404-9585-a567104a1a71"


@patch("sym.flow.cli.helpers.boto.uuid4")
@patch("sym.flow.cli.helpers.debug.put_public_fileobj")
def test_upload_terraform_output(
    put_public_fileobj_mock,
    uuid4_mock,
    click_setup,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    uuid4_mock.return_value = UUID(UUID_VAL)

    with click_setup() as runner:
        with capture_command():
            sandbox.create_binary("bin/terraform")
            dir = sandbox.create_dir("foo")
            capture_command.register_output("terraform", JSON)

            result = runner.invoke(
                click_command,
                ["debug", "upload-terraform-output", str(dir)],
            )
            assert result.exit_code == 0
            assert UUID_VAL in result.output

            put_public_fileobj_mock.assert_called_once()
            assert put_public_fileobj_mock.call_args.args[0].getvalue() == JSON.encode()
            assert UUID_VAL in put_public_fileobj_mock.call_args.args[1]



import factory

from sym.flow.cli.models.token import SymToken, SymTokenUser


class SymTokenUserFactory(factory.Factory):
    class Meta:
        model = SymTokenUser

    id = factory.Faker("uuid4")
    type = "normal"


class SymTokenFactory(factory.Factory):
    class Meta:
        model = SymToken

    identifier = factory.Faker("uuid4")
    user = factory.SubFactory(SymTokenUserFactory, type="bot")
    created_by = factory.SubFactory(SymTokenUserFactory)
    expires_at = factory.Faker("date_time")
    updated_at = factory.Faker("date_time")
    label = factory.Faker("word")



from pydantic import BaseModel, Field

SERVICE_KEY_SEPARATOR = ":"
SYM_CLOUD_KEY = "sym:cloud"


class BaseService(BaseModel):
    service_type: str = Field(alias="slug")
    external_id: str

    @property
    def service_key(self):
        return SERVICE_KEY_SEPARATOR.join([self.service_type, self.external_id])


class Service(BaseService):
    id: str
    label: str



from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.version import __version__


def test_version(click_setup):
    with click_setup() as runner:
        result = runner.invoke(click_command, ["version"])
        assert result.exit_code == 0
        assert result.output == f"{__version__}\n"



from unittest.mock import patch

from sym.flow.cli.helpers.config import store_login_config
from sym.flow.cli.symflow import symflow as click_command

EMAIL = "yasyf@symops.io"


def test_status_logged_out(click_setup):
    with click_setup() as runner:
        result = runner.invoke(click_command, ["status"])
        assert "are not currently logged in" in result.output
        assert result.exit_code == 1


@patch(
    "sym.flow.cli.helpers.api.SymAPI.verify_login",
    return_value=False,
)
def test_status_expired(mock_verify_login, click_setup, test_org, auth_token):
    with click_setup() as runner:
        store_login_config(EMAIL, test_org, auth_token)
        result = runner.invoke(click_command, ["status"])
        assert "Token expired" in result.output
        assert result.exit_code == 1


@patch(
    "sym.flow.cli.helpers.api.SymAPI.verify_login",
    return_value=True,
)
def test_status_success(mock_verify_login, click_setup, test_org, auth_token):
    with click_setup() as runner:
        store_login_config(EMAIL, test_org, auth_token)
        result = runner.invoke(click_command, ["status"])
        assert result.exit_code == 0



from unittest.mock import patch

import pytest

from sym.flow.cli.commands.users.list import get_user_data
from sym.flow.cli.helpers.utils import human_friendly, utc_to_local
from sym.flow.cli.tests.factories.users import UserFactory


class TestUsersList:
    @pytest.fixture
    def users(self):
        yield [
            UserFactory.create(),
            UserFactory.create(is_admin=True),
            UserFactory.create(),
        ]

    @pytest.fixture(autouse=True)
    def patchypatch(self, users):
        with patch("sym.flow.cli.helpers.api.SymAPI.get_users", return_value=users):
            yield

    def test_get_user_data(self, users, global_options):
        table_data = get_user_data(global_options.sym_api)

        assert table_data.pop(0) == ["Email", "Role", "Created At"]
        for i, row in enumerate(table_data):
            assert row == [
                users[i].sym_identifier,
                "admin" if users[i].is_admin else "",
                human_friendly(utc_to_local(users[i].created_at)),
            ]



from typing import List

import click
from tabulate import tabulate

from sym.flow.cli.errors import InvalidChoiceError
from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.models.resource import ResourceType


@click.command(name="list", short_help="List Sym resources")
@click.argument("resource_type", required=True, type=str)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def resources_list(options: GlobalOptions, resource_type: str) -> None:
    """Prints a table view of Sym resources in your Organization to STDOUT.

    Resources should be created and managed through Terraform
    (https://docs.symops.com/docs/terraform-provider).
    """

    resource_options = ResourceType.options()
    if resource_type not in resource_options:
        raise InvalidChoiceError(value=resource_type, valid_choices=resource_options)

    # We need to use getattr to retrieve the correct Enum item because MyEnum(value) uses the
    # value of the Enum item, not the name. E.g. ResourceType("SYM_FLOW") will fail but
    # ResourceType("flows") will succeed.
    resource_type_enum = getattr(ResourceType, resource_type.upper())
    table_data = get_resource_data(options.sym_api, resource_type_enum)
    if not table_data:
        click.echo(f"No {resource_type} resources were found.")
        return

    click.echo(tabulate(table_data, headers="firstrow"))


def get_resource_data(api: SymAPI, resource_type: ResourceType) -> List[List[str]]:
    resources = api.get_resources(resource_type)
    if not resources:
        return []

    has_sub_type = bool(resources[0].sub_type)
    headers = ["SRN", "Slug"]
    if has_sub_type:
        headers.append("Type")

    table_data = [headers]
    for resource in resources:
        resource_data = [resource.srn, resource.slug]
        if has_sub_type:
            resource_data.append(resource.sub_type)

        table_data.append(resource_data)

    return table_data



from sym.flow.cli.models.resource import ResourceType


class TestResource:
    def test_resource_type_options(self):
        options = ResourceType.options()
        assert "sym_flow" in options
        assert "sym_error_logger" in options



from io import IOBase
from pathlib import Path
from typing import Tuple
from urllib.parse import quote
from uuid import uuid4

import boto3
from sym.shared.cli.helpers.boto import make_intercept_boto_errors

from sym.flow.cli.helpers.constants import (
    DROPBOX_BUCKET,
    WRITE_ONLY_AWS_ACCESS_KEY_ID,
    WRITE_ONLY_AWS_SECRET_ACCESS_KEY,
)

intercept_boto_errors = make_intercept_boto_errors({})


def path_for_uploading(path: Path, ext: str) -> Tuple[str, str]:
    """Create a file name for the upload that includes the full
    path turned into hyphens so it is easy to find the right file.

    Will prefix the file name with a randomly generated uuid and return
    a tuple of that uuid and the full file name.
    """
    path_name = quote(str(path)).strip("/").replace("/", "-")
    unique_uuid = str(uuid4())
    return unique_uuid, str(Path(unique_uuid) / f"{path_name}.{ext}")


@intercept_boto_errors
def put_public_fileobj(file: IOBase, object_path: str, bucket: str = DROPBOX_BUCKET):
    s3_client = boto3.client(
        "s3",
        # Typically we don't include access keys like this, but are including write-only dropbox
        # keys here until we can generate temporary signed URLs via the API
        aws_access_key_id=WRITE_ONLY_AWS_ACCESS_KEY_ID,
        aws_secret_access_key=WRITE_ONLY_AWS_SECRET_ACCESS_KEY,
    )
    s3_client.upload_fileobj(
        file,
        bucket,
        object_path,
        ExtraArgs={"ACL": "bucket-owner-full-control"},
    )



import json
import uuid
from typing import Literal, Optional

import requests
from requests.exceptions import SSLError

from sym.flow.cli.errors import (
    NotAuthorizedError,
    SymAPIAggregateError,
    SymAPIRequestError,
    SymAPIUnknownError,
)

MethodT = Literal["GET", "HEAD", "PATCH", "POST", "PUT", "DELETE"]


class SymRESTClient:
    """Basic HTTP client for the Sym API."""

    last_request_id: str

    def __init__(self, url: str, access_token: Optional[str]):
        self.base_url = url
        self.access_token = access_token
        self.last_request_id = None  # type: ignore

    def make_headers(self, force_auth: bool = True) -> dict:
        if force_auth and not self.access_token:
            raise NotAuthorizedError()

        request_id = self.last_request_id = str(uuid.uuid4())
        headers = {
            "X-Sym-Request-ID": request_id,
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        if self.access_token:
            headers["Authorization"] = f"Bearer {self.access_token}"

        return headers

    def handle_response(
        self,
        response: requests.Response,
        *,
        validate: bool = True,
    ) -> requests.Response:
        """Validate a response from the server and raise, or return the response."""
        if not validate:
            return response

        if response.ok:
            # If the response was OK but didn't respond with any data,
            # don't try to parse it as JSON.
            if not response.content:
                return response

            try:
                response.json()
            except json.decoder.JSONDecodeError as err:
                raise SymAPIRequestError(
                    "The Sym API returned a malformed response.",
                    request_id=self.last_request_id,
                ) from err
            else:
                return response

        try:
            details = response.json()
            if error := details.get("message"):
                raise SymAPIAggregateError(error, self.last_request_id)
            elif errors := details.get("errors"):
                raise SymAPIAggregateError(errors, self.last_request_id)
            elif error := details.get("error"):
                raise SymAPIAggregateError(error, self.last_request_id)
            else:
                raise SymAPIUnknownError(
                    response_code=response.status_code,
                    request_id=self.last_request_id,
                )
        except (ValueError, KeyError):
            raise SymAPIUnknownError(
                response_code=response.status_code,
                request_id=self.last_request_id,
            )

    def _request(
        self,
        method: MethodT,
        endpoint: str,
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        validate: bool = True,
        force_auth: bool = True,
    ) -> requests.Response:
        try:
            response = requests.request(
                method,
                url=f"{self.base_url}/{endpoint}",
                params=params or {},
                data=json.dumps(data) if data else None,
                headers=self.make_headers(force_auth=force_auth),
            )
        except SSLError as err:
            raise SymAPIUnknownError(
                response_code=0, request_id=self.last_request_id
            ) from err

        return self.handle_response(response, validate=validate)

    def get(
        self,
        endpoint: str,
        params: Optional[dict] = None,
        *,
        validate: bool = True,
        force_auth: bool = True,
    ) -> requests.Response:
        return self._request(
            "GET", endpoint, params=params, validate=validate, force_auth=force_auth
        )

    def delete(self, endpoint: str, params: Optional[dict] = None) -> requests.Response:
        return self._request("DELETE", endpoint, params=params)

    def head(self, endpoint: str, params: Optional[dict] = None) -> requests.Response:
        return self._request("HEAD", endpoint, params=params)

    def patch(self, endpoint: str, data: Optional[dict] = None) -> requests.Response:
        return self._request("PATCH", endpoint, data=data)

    def post(self, endpoint: str, data: Optional[dict] = None) -> requests.Response:
        return self._request("POST", endpoint, data=data)

    def put(self, endpoint: str, data: Optional[dict] = None) -> requests.Response:
        return self._request("PUT", endpoint, data=data)



import click

from ...helpers.debug import do_upload_file


@click.command(short_help="Share a file with Sym")
@click.argument("file", type=click.Path(exists=True, dir_okay=False))
def upload_file(file: click.Path) -> None:
    """Share the contents of FILE with Sym."""

    click.echo("Uploading...")
    txid = do_upload_file(str(file))
    click.secho(f"Your file was shared with Sym!", fg="green")
    click.secho(f"Transaction ID: {txid}", fg="black", bold=True)



from unittest.mock import patch

import pytest

from sym.flow.cli.commands.services.services_list import get_services_data
from sym.flow.cli.errors import NotAuthorizedError
from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.conftest import get_mock_response
from sym.flow.cli.tests.helpers.test_api import MOCK_SERVICES_DATA

MOCK_SERVICES_OUTPUT_STR = """
Service Type    External ID    Label
--------------  -------------  -------
google          symops.io      Google
sym             cloud          Sym
""".strip()


class TestServicesList:
    @patch("sym.flow.cli.commands.services.services_list.click.echo")
    @patch(
        "sym.flow.cli.commands.services.services_list.get_services_data",
        return_value="real data",
    )
    def test_click_calls_execution_method(
        self,
        mock_get_services_data,
        mock_click_echo,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["services", "list"])
            assert result.exit_code == 0

        mock_get_services_data.assert_called_once()
        mock_click_echo.assert_called_once_with("real data")

    @patch(
        "sym.flow.cli.commands.services.services_list.get_services_data",
        side_effect=ValueError("random error"),
    )
    def test_click_call_catches_unknown_error(self, mock_get_services_data, click_setup):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["services", "list"])
            assert result.exit_code == 1
            assert isinstance(result.exception, ValueError)
            assert str(result.exception) == "random error"

        mock_get_services_data.assert_called_once()

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.get_services",
        side_effect=NotAuthorizedError,
    )
    def test_services_list_not_authorized_errors(self, mock_get_services, global_options):
        with pytest.raises(NotAuthorizedError, match="symflow login"):
            get_services_data(global_options.sym_api)
        mock_get_services.assert_called_once()

    @patch(
        "sym.flow.cli.helpers.api.SymRESTClient.get",
        return_value=(get_mock_response(200, data=MOCK_SERVICES_DATA)),
    )
    def test_services_list(self, mock_get_services, global_options):
        data = get_services_data(global_options.sym_api)
        assert data == MOCK_SERVICES_OUTPUT_STR



import os

import pytest
import yaml
from sym.shared.cli.helpers.contexts import push_env

from ...helpers.config import Config, deepget, store_login_config
from ...models import Organization


@pytest.fixture
def org() -> Organization:
    return Organization(slug="slug", client_id="client-id")


@pytest.fixture
def email() -> str:
    return "ci@symops.io"


__config_yaml = """
auth_token:
  access_token: access
  expires_in: 86400
  refresh_token: refresh
  scope: scopes
  token_type: type
client_id: client-id
email: ci@symops.io
org: slug
"""


def test_write_login_config(sandbox, org, email, auth_token):
    with push_env("XDG_CONFIG_HOME", str(sandbox.path / ".config")):
        path = store_login_config(email=email, org=org, auth_token=auth_token)
        with open(path) as fd:
            data = fd.read()
            assert data.strip() == __config_yaml.strip()


def test_read_login_config(sandbox, org, email, auth_token):
    with push_env("XDG_CONFIG_HOME", str(sandbox.path / ".config")):
        config_dir = sandbox.path / ".config" / "symflow" / "default"
        os.makedirs(config_dir)
        with open(str(config_dir / "config.yml"), "w") as fd:
            fd.write(__config_yaml)
            fd.flush()
            assert Config.get_email() == email
            assert Config.get_org() == org
            assert Config.get_access_token() == "access"


def test_logout(sandbox, org, email, auth_token):
    with push_env("XDG_CONFIG_HOME", str(sandbox.path / ".config")):
        path = store_login_config(email=email, org=org, auth_token=auth_token)
        with open(path) as fd:
            data = fd.read()
            assert data.strip() == __config_yaml.strip()

        Config.logout()
        with open(path, "r") as fd:
            data = yaml.safe_load(fd.read())
            assert "email" not in data


def test_config_get_org(org, email, auth_token):
    store_login_config(email=email, org=org, auth_token=auth_token)
    assert Config.get_value("org") == org.slug


TEST_DICT = {"first": {"second": {"foo": "bar"}}}


def test_deepget_should_return_none_on_bad_key():
    assert deepget("wrong.bad", TEST_DICT) is None


def test_deepget_should_return_the_correct_value():
    assert deepget("first.second.foo", TEST_DICT) == "bar"



import webbrowser

import click

from sym.flow.cli.helpers.api import SymAPI


def slack_new(api: SymAPI, service_id: str) -> None:
    """Generate a magic link to install the Sym Slack App. This link
    can be opened directly or sent to an administrator with permission to
    install the app into your workspace.
    """
    url = get_magic_url(api=api, service_id=service_id)
    click.echo("Generated an installation link for the Sym Slack App:\n")
    click.secho(url, fg="white", bold=True)
    click.echo(
        "\nPlease send this URL to an administrator who has permission to install the app.\nOr, if that's you, we can open it now."
    )
    if click.confirm(
        "\nWould you like to open the Slack installation URL in a browser window?",
        default=True,
    ):
        webbrowser.open(url)


def get_magic_url(api: SymAPI, service_id: str):
    return api.get_slack_install_url(service_id)



import factory

from sym.flow.cli.models.resource import TerraformResource


class TerraformResourceFactory(factory.Factory):
    class Meta:
        model = TerraformResource

    id = factory.Faker("uuid4")
    slug = factory.Sequence(lambda n: "resource-slug-%03d" % n)
    type = "slack"
    srn = factory.Sequence(
        lambda n: f"test_org:sym_integration:slack:resource-slug-%03d" % n
    )



from contextlib import contextmanager
from typing import Optional

from sym.shared.cli.helpers.config.base import ConfigBase
from sym.shared.cli.tests.helpers.sandbox import Sandbox as SharedSandbox

from sym.flow.cli.helpers.config import Config


class Sandbox(SharedSandbox):
    def reset_config(self):
        ConfigBase.reset()
        Config.reset()

    @contextmanager
    def setup(self, set_org: Optional[bool] = True, set_client: Optional[bool] = True):
        with self.push_xdg_config_home():
            with self.push_exec_path():
                yield



import pytest

from sym.flow.cli.models.service import BaseService
from sym.flow.cli.models.user import CSVMatcher, Identity


class TestCSVMatcher:
    def test_dict_init(self):
        matcher = {"email": "user@symops.io"}
        service = BaseService(slug="sym", external_id="cloud")
        csv_matcher = CSVMatcher.from_dict(service=service, matcher=matcher)
        assert csv_matcher.value == "user@symops.io"
        assert csv_matcher.to_dict() == matcher

    def test_unknown_mapping(self):
        service = BaseService(slug="twitter", external_id="cloud")
        value = '{"some_unknown_service":"user@sus.io"}'
        matcher = CSVMatcher(service=service, value=value)
        assert matcher.to_dict() == {"some_unknown_service": "user@sus.io"}

    def test_invalid_unknown_mapping(self):
        service = BaseService(slug="twitter", external_id="cloud")
        value = "this is not json"
        matcher = CSVMatcher(service=service, value=value)
        with pytest.raises(ValueError, match="should be valid json"):
            matcher.to_dict() == {"some_unknown_service": "user@sus.io"}


class TestIdentity:
    def test_regular_init(self):
        matcher = {"email": "user@symops.io"}
        service = BaseService(slug="google", external_id="symops.io")
        identity = Identity(matcher=matcher, service=service)
        assert identity.to_csv() == matcher["email"]

    def test_from_csv(self):
        matcher = {"email": "user@symops.io"}
        identity = Identity.from_csv(
            service_key="google:symops.io", matcher_value=matcher["email"]
        )
        assert identity.to_csv() == matcher["email"]

    def test_parse_service_key(self):
        service_type, external_id = Identity.parse_service_key("google:symops.io")
        assert service_type == "google"
        assert external_id == "symops.io"

    def test_equality(self):
        assert Identity.parse_service_key(
            "slack:symops.io"
        ) != Identity.parse_service_key("google:symops.io")
        assert Identity.parse_service_key(
            "google:symops.io"
        ) == Identity.parse_service_key("google:symops.io")



from typing import Optional

import click
import requests
from semver import VersionInfo
from sentry_sdk import capture_exception

from sym.flow.cli.version import __version__

from .constants import CLI_PACKAGE_NAME


def get_current_version() -> VersionInfo:
    """Returns the current version of the CLI."""

    try:
        return VersionInfo.parse(__version__)
    except ValueError as e:
        # Log the error in Sentry but don't crash the CLI.
        capture_exception(e)
        return VersionInfo.parse("0.0.0")


def get_latest_version() -> Optional[VersionInfo]:
    """Retrieves the latest version of the CLI from PyPI."""

    try:
        response = requests.get(f"https://pypi.org/pypi/{CLI_PACKAGE_NAME}/json").json()
    except Exception as e:
        # Log the error in Sentry but don't crash the CLI.
        capture_exception(e)
        return None

    try:
        return VersionInfo.parse(response["info"]["version"])
    except (KeyError, ValueError) as e:
        # Log the error in Sentry but don't crash the CLI.
        capture_exception(e)
        return None


def maybe_display_update_message() -> None:
    """If there is a newer version of the CLI available, displays a message to
    the user. Otherwise, does nothing.
    """

    if (
        latest_version := get_latest_version()
    ) and latest_version > get_current_version():
        click.secho(
            f"Your version of symflow is out of date! The latest version is {latest_version}.\n"
            "If you're using Homebrew, you can update by running `brew update && brew upgrade symopsio/tap/symflow`.\n"
            "Otherwise, please see https://docs.symops.com/docs/install-sym-flow-cli",
            fg="yellow",
            err=True,
        )



from unittest.mock import call, patch

import pytest
from cli.helpers.utils import filter_dict, get_or_prompt

from sym.flow.cli.errors import InvalidChoiceError


@pytest.fixture
def mock_prompt():
    yield "Pick one"


def test_filter_dict():
    d = {"a": 1, "b": 2}

    filtered_d = filter_dict(d, lambda v: v > 1)
    assert filtered_d == {"b": 2}


@patch("inquirer.list_input")
def test_get_or_prompt_invokes_inquirer(mock_inquire, mock_prompt):
    get_or_prompt(value=None, prompt=mock_prompt, choices=["foo", "bar"])
    # choices are sorted before invoking inquire
    mock_inquire.assert_has_calls([call(mock_prompt, choices=["bar", "foo"])])


@patch("click.echo")
def test_get_or_prompt_auto_selects(mock_echo, mock_prompt):
    # value = None, but only one choice available. Auto selects
    result = get_or_prompt(value=None, prompt=mock_prompt, choices=["foo"])
    assert result == "foo"
    mock_echo.assert_has_calls([call(f"{mock_prompt}: Using 'foo'")])


def test_get_or_prompt_invalid_choice(mock_prompt):
    with pytest.raises(InvalidChoiceError):
        get_or_prompt(value="bad", prompt=mock_prompt, choices=["foo"])


def test_get_or_prompt(mock_prompt):
    # value exists and is valid
    assert get_or_prompt(value="foo", prompt=mock_prompt, choices=["foo", "bar"]) == "foo"



from typing import List, Optional

import click
import inquirer

from sym.flow.cli.errors import MissingChoicesError


class InquirerChoiceOption(click.Option):
    """Custom click.Option that uses inquirer to prompt."""

    def __init__(self, *args, **kwargs):
        self.message: Optional[str] = kwargs.pop("inquirer_prompt", None)
        try:
            self.choices: List = kwargs.pop("inquirer_choices")
        except KeyError:
            raise MissingChoicesError()

        super().__init__(*args, **kwargs)

    def prompt_for_value(self, ctx):
        """Replaces click prompt with inquirer's prompt.
        Prompts for the given service types"""
        value = inquirer.list_input(
            self.message or self.prompt, choices=sorted(self.choices)
        )
        return self.process_value(ctx, value)



import uuid
from unittest.mock import patch

from sym.flow.cli.symflow import symflow as click_command


@patch(
    "sym.flow.cli.helpers.api.SymAPI.revoke_token",
)
class TestTokensRevoke:
    def test_revoke_token(self, mock_api, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["tokens", "revoke", str(uuid.uuid4())],
            )
            assert result.exit_code == 0



from typing import List
from unittest.mock import patch
from uuid import UUID

import pytest

from sym.flow.cli.commands.tokens.list import get_tokens_data, get_user
from sym.flow.cli.helpers.utils import human_friendly, utc_to_local
from sym.flow.cli.models import User
from sym.flow.cli.tests.factories.tokens import SymTokenFactory
from sym.flow.cli.tests.factories.users import UserFactory


class TestTokensList:
    @pytest.fixture
    def tokens(self):
        yield SymTokenFactory.create_batch(3)

    @pytest.fixture
    def users(self):
        yield UserFactory.create_batch(2)

    def test_get_tokens_data(self, tokens, users, global_options):
        with patch(
            "sym.flow.cli.helpers.api.SymAPI.get_tokens",
            return_value=tokens,
        ):
            # API will be called twice per token, because factory generates 6 users
            with patch(
                "sym.flow.cli.helpers.api.SymAPI.get_user_by_id",
                side_effect=[users[0], users[1]] * 3,
            ):
                table_data = get_tokens_data(global_options.sym_api)

        assert table_data.pop(0) == [
            "Token ID",
            "User",
            "Created By",
            "Expires At",
            "Last Used",
            "Label",
        ]
        for i, row in enumerate(table_data):
            assert row == [
                tokens[i].identifier,
                users[0].sym_identifier,
                users[1].sym_identifier,
                human_friendly(utc_to_local(tokens[i].expires_at)),
                human_friendly(utc_to_local(tokens[i].updated_at)),
                tokens[i].label,
            ]

    def test_get_user(self, users: List[User], global_options):
        user_cache = {}
        with patch(
            "sym.flow.cli.helpers.api.SymAPI.get_user_by_id",
            side_effect=[users[0], users[1]],
        ) as mock_sym_api_get:
            # This should return side_effect 1 of the patch
            assert (
                get_user(UUID(users[0].id), user_cache, global_options.sym_api)
                == users[0]
            )

            # API should not be called. We will assert the number of calls at the end
            assert (
                get_user(UUID(users[0].id), user_cache, global_options.sym_api)
                == users[0]
            )

            # This should return side_effect 2 of the patch
            assert (
                get_user(UUID(users[1].id), user_cache, global_options.sym_api)
                == users[1]
            )

            # only called twice, once per unique user.
            assert mock_sym_api_get.call_count == 2
